# Copyright 2024, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT
from __future__ import annotations

from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    Iterable,
    Iterator,
    Protocol,
    Sequence,
    Tuple,
    TypeVar,
    runtime_checkable,
)

from typing_extensions import TypeAlias

# Note
# (1) the use of each generic variable can differ in generic components
# (2) the use of a generic affects the statically-correct way to declare that generic's type variance
# Because of (1) and (2), I can't use the same TypeVars in all places where an input/target/metadata object
# is expected. This sacrifices some clarity in this file, but hopefully the classes can appear intuitive
# from the perspective of the end-users that only interact with parameterized generics.

# (3) Methods/signatures advertised by a protocol class *must* be mirrored by
# compatible types in implementation. If overload decorator is present,
# only the 'overload'-decorated methods are the source of these "promised" signatures
# for the type checker. If more than one signature is advertised by a protocol,
# then implementers must use 'overload'-decorator to advertise compatible signatures.

# In light of these rules, what should we do?
#   - Including overload in protocol signatures is clear, but does require
#     protocol implementers to leverage 'overload' decorator, which isn't quite 'beginner' level
#   - Not including overload in protocol signatures makes it possible to implement
#     protocols without helping type checker determine specific return types.
#     this is more flexible for implementer, but will require more type narrowing
#     on non-specific ("union") return types. (If typechecker can't determine return
#     type from Model('BatchInputType') is a 'BatchTargetType' and not an TargetType,
#     user has to type narrow.
#   - Seems like overload decorator is the cost of handling multiple types of input arguments


# create instance-level versions of all generic type vars
InputType_co = TypeVar("InputType_co", covariant=True)
TargetType_co = TypeVar("TargetType_co", covariant=True)
DatumMetadataType_co = TypeVar("DatumMetadataType_co", covariant=True)

InputType_cn = TypeVar("InputType_cn", contravariant=True)
TargetType_cn = TypeVar("TargetType_cn", contravariant=True)
DatumMetadataType_cn = TypeVar("DatumMetadataType_cn", contravariant=True)

InputType_in = TypeVar("InputType_in", contravariant=False, covariant=False)
TargetType_in = TypeVar("TargetType_in", contravariant=False, covariant=False)
DatumMetadataType_in = TypeVar(
    "DatumMetadataType_in", contravariant=False, covariant=False
)


# create batch "versions" of all type vars
InputBatchType_co = TypeVar("InputBatchType_co", covariant=True)
TargetBatchType_co = TypeVar("TargetBatchType_co", covariant=True)
DatumMetadataBatchType_co = TypeVar("DatumMetadataBatchType_co", covariant=True)

InputBatchType_cn = TypeVar("InputBatchType_cn", contravariant=True)
TargetBatchType_cn = TypeVar("TargetBatchType_cn", contravariant=True)
DatumMetadataBatchType_cn = TypeVar("DatumMetadataBatchType_cn", contravariant=True)

InputBatchType_in = TypeVar("InputBatchType_in", contravariant=False, covariant=False)
TargetBatchType_in = TypeVar("TargetBatchType_in", covariant=False, contravariant=False)
DatumMetadataBatchType_in = TypeVar(
    "DatumMetadataBatchType_in", covariant=False, contravariant=False
)

MetricComputeReturnType = Dict[str, Any]

# TODO 0: add docstrings
#
# TODO 1: check type expected by pytorch getitem
#
# TODO 2: Decide if I really need different TypeVars based on context?
#           - in the context of Dataset, class is covariant to all types
#           - in the context of Augmentation, class is invariant to all types
# TODO 3: Consider whether 'to' method should be a part of Metric/Model/Dataset protocol
#           (intuitively, this seems more like framework-specific implementation detail)
#
# TODO 4: Consider whether using Datum as a TypeAlias is more confusing than helpful
#         It seems we just need 3 typevars in the TypeAlias assignment and the TypeVar
#         type variance is completely inconsequential (because they are substituted for
#         when the TypeAlias is used (as 'Any' or as the provided bracketed types.) We
#         could also define a 4th version of Input/Target/Metadata types, but this also
#         seems confusing.
# TODO 5: Consider how easily and usefully variadic generics (which sound a bit scary)
#         could be used to helpfully represent the shape of an expected array.
#         So, for example, instead of having type hints that specified that 'InputType=ArrayLike'
#         we could say 'InputType=ArrayLike[H,W,C]'.
#
# TODO 6: Add AugmentationMetadata
#
# TODO 7: Verify use of 'overload' decorator in protocol definition
#       Methods/signatures advertised by a protocol class *must* be mirrored by
#       compatible types in implementation. If overload decorator is present,
#       only the overloaded methods are the source of these "promised" signatures.
#       If more than one signature is advertised by a protocol, then implementers
#       must use overload-decorator to advertise compatible signatures.


# Generic versions of all protocols

# myee: this version of Dataset was identical to DataLoader so both redundant
# - and made it hard to distinguish between the two classes

# class Dataset(Protocol, Generic[InputType_co, TargetType_co, DatumMetadataType_co]):
#     def __iter__(
#             self
#     ) -> Iterator[Tuple[InputType_co, TargetType_co, DatumMetadataType_co]]:
#         ...


@runtime_checkable
class Dataset(Protocol, Generic[InputType_co, TargetType_co, DatumMetadataType_co]):
    def __getitem__(
        self, __ind: int
    ) -> Tuple[InputType_co, TargetType_co, DatumMetadataType_co]:
        ...

    def __len__(self) -> int:
        ...


@runtime_checkable
class DataLoader(Protocol, Generic[InputType_co, TargetType_co, DatumMetadataType_co]):
    def __iter__(
        self,
    ) -> Iterator[Tuple[InputType_co, TargetType_co, DatumMetadataType_co]]:
        ...

    # no longer having Dataloader operate as the Iterator, just using it as an iterable
    # so it doesn't need to return itself from the __iter__ method nor have a __next__ method


@runtime_checkable
class Model(
    Protocol,
    Generic[InputBatchType_cn, TargetBatchType_co],
):
    def __call__(self, __batch_input: InputBatchType_cn) -> TargetBatchType_co:
        ...


@runtime_checkable
class Metric(Protocol, Generic[TargetBatchType_cn]):
    def reset(self) -> None:
        ...

    def update(
        self,
        __preds_batch: TargetBatchType_cn,
        __targets_batch: TargetBatchType_cn,
    ) -> None:
        ...

    def compute(self) -> MetricComputeReturnType:
        ...

    # don't believe Metric needs to guarantee a 'to' method exists
    # if all inputs to update are on GPU, operation should happen in framework specific way


@runtime_checkable
class Augmentation(
    Protocol,
    Generic[
        InputBatchType_co,
        TargetBatchType_co,
        DatumMetadataBatchType_co,
        InputBatchType_cn,
        TargetBatchType_cn,
        DatumMetadataBatchType_cn,
    ],
):
    def __call__(
        self,
        __batch: Tuple[
            InputBatchType_cn, TargetBatchType_cn, DatumMetadataBatchType_cn
        ],
    ) -> Tuple[InputBatchType_co, TargetBatchType_co, DatumMetadataBatchType_co]:
        ...


T_in = TypeVar("T_in")
T_tgt = TypeVar("T_tgt")
T_md = TypeVar("T_md")

CollateFn: TypeAlias = Callable[
    [Iterable[Tuple[T_in, T_tgt, T_md]]],
    Tuple[Sequence[T_in], Sequence[T_tgt], Sequence[T_md]],
]



def person(name: str, age: int) -> None:
    """
    Enter person ID info.

    Parameters
    ----------
    name : str
        The person's first name.
    age : int
        The person's age.

    Returns
    -------
    None
        Nothing.

    Examples
    --------
    >>> from maite.testing.documentation.documentation_dependencies import person
    >>> person('Brad', 22)
    """

    return



# Copyright 2024, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT

from pathlib import Path

from maite.testing.project import ModuleScan

module_scan = ModuleScan()

all_dummy_subpkgs = sorted(
    p.name for p in (Path(__file__).parent / "dummy_projects").glob("*") if p.is_dir()
)

assert all_dummy_subpkgs



# Copyright 2024, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT

import importlib
import importlib.util


def is_numpy_available():
    return importlib.util.find_spec("numpy") is not None


def is_torch_available():
    return importlib.util.find_spec("torch") is not None


def is_torchvision_available():
    return importlib.util.find_spec("torchvision") is not None


def is_timm_available():
    return importlib.util.find_spec("timm") is not None


def is_hf_hub_available():
    return importlib.util.find_spec("huggingface_hub") is not None


def is_hf_datasets_available():
    return importlib.util.find_spec("datasets") is not None


def is_hf_transformers_available():
    return importlib.util.find_spec("transformers") is not None


def is_pytest_available():
    return importlib.util.find_spec("pytest") is not None


def is_torchmetrics_available():
    return importlib.util.find_spec("torchmetrics") is not None


def is_tqdm_available():
    return importlib.util.find_spec("tqdm") is not None


def is_pil_available():
    return importlib.util.find_spec("PIL") is not None


def is_torcheval_available():
    return importlib.util.find_spec("torcheval") is not None


def is_hypothesis_available():
    return importlib.util.find_spec("hypothesis") is not None


if is_pytest_available():
    import pytest

    requires_torch = pytest.mark.skipif(
        not is_torch_available(), reason="test requires torch"
    )
    requires_torchvision = pytest.mark.skipif(
        not is_torchvision_available(), reason="test requires torchvision"
    )
    requires_timm = pytest.mark.skipif(
        not is_timm_available(), reason="test requires timm"
    )
    requires_hf_hub = pytest.mark.skipif(
        not is_hf_hub_available(), reason="test requires huggingface_hub"
    )
    requires_hf_datasets = pytest.mark.skipif(
        not is_hf_datasets_available(), reason="test requires datasets"
    )
    requires_hf_transformers = pytest.mark.skipif(
        not is_hf_transformers_available(), reason="test requires transformers"
    )
    requires_torchmetrics = pytest.mark.skipif(
        not is_torchmetrics_available(), reason="test requires torchmetrics"
    )
    requires_pil = pytest.mark.skipif(
        not is_pil_available(), reason="test requires PIL"
    )
    requires_torcheval = pytest.mark.skipif(
        not is_torcheval_available(), reason="test requires torcheval"
    )



# Copyright 2024, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT

# flake8: noqa

import os
import subprocess
import sys
from importlib.util import find_spec
from pathlib import Path

import hypothesis.strategies as st
from hypothesis import settings

from maite._internals import import_utils
from maite.testing.pytest import cleandir  # noqa: F401
from tests import all_dummy_subpkgs

st.register_type_strategy(st.DataObject, st.data())

settings.register_profile("cicd", max_examples=10, deadline=None)

if bool(os.environ.get("CI_JOB_ID")):
    print("*** Running in CI, using CI settings ***")
    settings.load_profile("cicd")

# Skip collection of tests that don't work on the current version of Python.
collect_ignore_glob = []


def _safe_find_spec(pkg: str):
    # returning `None` means that module/subpackage is not installed
    try:
        return find_spec(pkg)
    except ModuleNotFoundError:
        return None


for subpkg in all_dummy_subpkgs:
    if _safe_find_spec(f"maite_dummy.{subpkg}") is None:
        subprocess.check_call(
            [
                sys.executable,
                "-m",
                "pip",
                "install",
                str((Path(__file__).parent / f"dummy_projects/{subpkg}").absolute()),
            ]
        )

if import_utils.is_torch_available():
    from hypothesis import register_random

    class TorchRandomWrapper:
        def __init__(self):
            # This class provides a shim that matches torch to stdlib random,
            # and lets us avoid importing torch until it's already in use.
            from torch import default_generator

            self.seed = default_generator.manual_seed
            self.getstate = default_generator.get_state
            self.setstate = default_generator.set_state

    r = TorchRandomWrapper()
    # Note: do not specifying TorchRandomWrapper() inline. It will be garbage collected
    register_random(r)
else:
    collect_ignore_glob.append("requires_torch/**")

if not import_utils.is_torch_available():
    collect_ignore_glob.append("*torch*.py")

if not (
    import_utils.is_hf_transformers_available()
    and import_utils.is_hf_datasets_available()
    and import_utils.is_hf_hub_available()
):
    collect_ignore_glob.append("*huggingface*.py")

if not import_utils.is_torchvision_available():
    collect_ignore_glob.append("*torchvision*.py")

if not import_utils.is_torchmetrics_available():
    collect_ignore_glob.append("*torchmetrics*.py")

if not import_utils.is_torcheval_available():
    collect_ignore_glob.append("*torcheval*.py")



# Copyright 2024, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT

from abc import ABCMeta, abstractmethod
from functools import wraps
from typing import Any, Callable, Dict, Iterable, Tuple, TypeVar, cast
from weakref import WeakSet

from maite._internals.import_utils import is_torch_available, is_tqdm_available
from maite._internals.protocols.generic import (
    DataLoader,
    DatumMetadataType_co,
    InputType_co,
    TargetType_co,
)

T = TypeVar("T", bound=Callable)


def is_typed_dict(obj: Any) -> bool:
    if not isinstance(obj, type):
        return False

    return all(
        hasattr(obj, attr)
        for attr in ("__required_keys__", "__optional_keys__", "__optional_keys__")
    )


class ContextDecorator(metaclass=ABCMeta):
    @abstractmethod
    def __enter__(self):  # pragma: no cover
        raise NotImplementedError()

    @abstractmethod
    def __exit__(self, type, value, traceback):  # pragma: no cover
        raise NotImplementedError()

    def __call__(self, func: T) -> T:
        @wraps(func)
        def wrapper(*args, **kwargs):
            with self:
                return func(*args, **kwargs)

        return cast(T, wrapper)


class evaluating(ContextDecorator):
    """A context manager / decorator that temporarily places one
    or more modules in eval mode during the context."""

    def __init__(self, *modules: Callable) -> None:
        """
        Parameters
        ----------
        *modules: Module

        Notes
        -----
        A module's state is restored faithfully; e.g., a module that
        was already in eval mode will not be placed in train mode upon
        leaving the `evaluating` context.

        Examples
        --------
        >>> from torch.nn import Linear
        >>> from maite.utils import evaluating

        Using `evaluating` as a context manager.

        >>> module = Linear(1, 1)
        >>> module.training
        True
        >>> with evaluating(module):
        ...     print(module.training)
        False
        >>> module.training
        True

        Using `evaluating` as a decorator.

        >>> def f():
        ...     print("hello world")
        ...     return module.training
        >>> f = evaluating(module)(f)
        >>> module.training
        True
        >>> f()
        hello world
        False
        >>> module.training
        True
        """
        self._states: Dict[bool, WeakSet[Callable]] = {
            True: WeakSet(),
            False: WeakSet(),
        }

        for m in modules:
            if is_torch_available():
                from torch.nn import Module

                if isinstance(m, Module):
                    self._states[m.training].add(m)

    def __enter__(self) -> None:
        for train_status in self._states:
            for m in self._states[train_status]:
                if is_torch_available():
                    from torch.nn import Module

                    if isinstance(m, Module):
                        m.eval()

    def __exit__(self, type, value, traceback) -> None:
        for train_status in self._states:
            for m in self._states[train_status]:
                if is_torch_available():
                    from torch.nn import Module

                    if isinstance(m, Module):
                        m.train(train_status)


def add_progress_bar(
    dataloader: DataLoader[InputType_co, TargetType_co, DatumMetadataType_co],
) -> Iterable[Tuple[InputType_co, TargetType_co, DatumMetadataType_co]]:
    """Wrap a dataloader with tqdm to display progress bars.

    Note tqdm output can be disabled as of entirely as of version 4.66.0 by setting the
    environment variable TQDM_DISABLE=1.

    Parameters
    ----------
    dataloader : DataLoader[InputType_co, TargetType_co, DatumMetadataType_co]
        The dataloader to wrap.

    Returns
    -------
    Iterable[Tuple[InputType_co, TargetType_co, DatumMetadataType_co]]
        Return an iterator over batches of data, where each batch contains a tuple of
        of model input, model target , and datum-level metadata.
    """
    if is_tqdm_available():
        # tqdm.auto will resolve to tqdm.autonotebook, tqdm.asyncio or tqdm.std
        # depending on the environment
        from tqdm.auto import tqdm

        return tqdm(dataloader)
    else:
        return dataloader



# Copyright 2024, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT

from itertools import chain
from typing import Optional, Protocol

import pytest
from pytest import param
from typing_extensions import TypedDict

from maite._internals.testing.docs import _get_numpy_tags
from maite.errors import InvalidArgument
from maite.testing.docs import validate_docstring


def form_doc(
    summary: str = "",
    extended: str = "",
    params: str = "",
    returns: str = "",
    see_also: str = "",
    examples: str = "",
):
    doc = "\n"
    if summary:
        doc += summary + "\n\n"
    if extended:
        doc += extended + "\n"
    if params:
        doc += f"""
Parameters
----------
{params}
"""

    if returns:
        doc += f"""
Returns
-------
{returns}
"""

    if see_also:
        doc += f"""
See Also
--------
{see_also}
"""
    if examples:
        doc += f"""
Examples
--------
{examples}
"""
    if doc.startswith("\n\n"):
        doc = doc[1:]
    return doc


def make_func(
    summary: str = "",
    extended: str = "",
    params: str = "",
    returns: str = "",
    see_also: str = "",
    examples: str = "",
):
    def f(x, y):
        ...

    f.__doc__ = form_doc(summary, extended, params, returns, see_also, examples)
    return f


def make_class(
    class_doc: Optional[str] = None,
    init_doc: Optional[str] = None,
    method_doc: str = "",
    property_doc: str = "",
):
    class Class:
        if init_doc is not None:

            def __init__(self, x: int) -> None:
                ...

        if method_doc:

            def method(self, y: str):
                ...

        if property_doc:

            @property
            def prop(self):
                ...

    if class_doc is not None:
        Class.__doc__ = class_doc

    if init_doc is not None:
        Class.__init__.__doc__ = init_doc

    if method_doc:
        Class.method.__doc__ = method_doc
    if property_doc:
        Class.prop.__doc__ = property_doc
    return Class


@pytest.mark.parametrize(
    "obj",
    [
        make_func(
            "Compute result.",
            "Uses math to do thing.",
            params="x : int\n    About x.\ny : str\n    About y.",
            returns="int\n    The result.",
            examples=">>> 1 + 1\n2",
        ),
        make_class(
            form_doc(
                "A class thing.",
                "This class does things.\nIt does lots of things.",
                "",
                examples=">>> 1+1\n2",
            ),
            form_doc(params="x : int\n    About x."),
        ),
        make_class(
            init_doc=form_doc(
                "A class thing.",
                "This class does things.\nIt does lots of things.",
                params="x : int\n    About x.",
                examples=">>> 1+1\n2",
            ),
        ),
        make_class(
            form_doc(
                "A class thing.",
                "This class does things.\nIt does lots of things.",
                params="x : int\n    About x.",
                examples=">>> 1+1\n2",
            ),
            init_doc="",
        ),
        make_class(
            form_doc(
                "A class thing.",
                "This class does things.\nIt does lots of things.",
                examples=">>> 1+1\n2",
            ),
        ),
        make_class(
            form_doc(
                examples=">>> 1+1\n2",
            ),
            form_doc(
                "A class thing.",
                "This class does things.\nIt does lots of things.",
                params="x : str\n    About x.",
            ),
        ),
        make_class(
            "",
            form_doc(
                "A class thing.",
                "This class does things.\nIt does lots of things.",
                params="x : str\n    About x.",
                examples=">>> 1+1\n2",
            ),
        ),
        make_class(
            form_doc(
                "A class thing.",
                "This class does things.\nIt does lots of things.",
                params="x : str\n    About x.",
                examples=">>> 1+1\n2",
            ),
            "",
        ),
        make_class(
            form_doc(
                "A class thing.",
                "This class does things.\nIt does lots of things.",
                params="x : str\n    About x.",
                examples=">>> 1+1\n2",
            ),
            "",
            method_doc=form_doc(
                "A class thing.",
                "This class does things.\nIt does lots of things.",
                params="y : str\n    About x.",
                examples=">>> 1+1\n2",
            ),
        ),
        pytest.param(
            make_class(),
            marks=pytest.mark.xfail(raises=AssertionError, strict=True),
            id="BadClass",
        ),
        pytest.param(
            make_func(),
            marks=pytest.mark.xfail(raises=AssertionError, strict=True),
            id="bad_func",
        ),
    ],
)
def test_good_doc(obj):
    results = validate_docstring(obj, ignore=("SA01",))
    assert results["error_count"] == 0, results["errors"]


bad_doc_func = make_func(
    "Compute result.",
    "Uses math to do thing.",
    params="z : int\n    About x.\ny : str\n    About y.",
    returns="int\n    The result.",
)

bad_doc_class = make_class(
    form_doc(
        "A class thing.",
        "This class does things.\nIt does lots of things.",
        "",
        examples=">>> 1+1\n2",
    ),
    form_doc(params="x : str\n    About x."),
)


@pytest.mark.parametrize(
    "obj, ignore_codes, error_codes",
    [
        param(bad_doc_func, ["SA01"], ["EX01", "PR01", "PR02"], id="check_ignore0"),
        param(bad_doc_func, ["EX01", "SA01"], ["PR01", "PR02"], id="check_ignore1"),
        param(bad_doc_func, ["EX01", "SA01", "PR01"], ["PR02"], id="check_ignore2"),
        param(bad_doc_func, ["EX01", "SA01", "PR01", "PR02"], [], id="check_ignore3"),
        param(
            make_class(
                form_doc(
                    "A class thing.",
                    "This class does things.\nIt does lots of things.",
                    "",
                    examples=">>> 1+1\n2",
                ),
                form_doc(params="x : str\n    About x."),
            ),
            ["SA01"],
            [],
            id="ignore extended summary",
        ),
        param(
            make_class(
                form_doc(
                    "A class thing.",
                    "This class does things.\nIt does lots of things.",
                ),
                form_doc(params="x : str\n    About x."),
            ),
            ["SA01", "EX01"],
            [],
            id="ignore examples",
        ),
        param(
            make_class(
                form_doc(
                    "A class thing.",
                    "This class does things.\nIt does lots of things.",
                ),
                "",
            ),
            ["SA01", "EX01", "PR01"],
            [],
            id="ignore params and examples",
        ),
        param(
            make_class(
                form_doc(
                    "A class thing.",
                    "This class does things.\nIt does lots of things.",
                    params="x : str\n    About x.",
                    examples=">>> 1+1\n2",
                )
            ),
            ["SA01"],
            ["PR02"],
            id="error: documented param x is unknown",
        ),
        param(make_class(), [], ["GL08"], id="error: No class docstring"),
    ],
)
def test_bad_doc(obj, ignore_codes, error_codes):
    results = validate_docstring(obj, ignore=ignore_codes, include_ignored_errors=True)
    assert set(results["errors"]) == set(error_codes), results["errors"]
    assert set(results["ignored_errors"]) == set(ignore_codes)
    assert results["error_count"] == sum(len(v) for v in results["errors"].values())


def test_bad_error_code():
    with pytest.raises(InvalidArgument, match=r"NOTACODE"):
        validate_docstring(make_class(), ignore=["NOTACODE"])  # type: ignore


def test_ignore_method():
    class_bad_method = make_class(
        method_doc=form_doc(
            "A class thing.",
            "This class does things.\nIt does lots of things.",
            params="y : str\n    About x.",
        ),
    )
    results1 = validate_docstring(class_bad_method, method_ignore=["EX01", "SA01"])
    assert results1["error_count"] == 1, results1["errors"]

    results2 = validate_docstring(class_bad_method)
    assert results2["error_count"] == 2
    assert "GL08" in results2["errors"]
    del results2["errors"]["GL08"]
    assert all(
        msg.startswith("Class.method") for msg in chain(*results2["errors"].values())
    )


def test_ignore_property():
    class_bad_method = make_class(
        property_doc=form_doc(
            "A class thing.",
            "This class does things.\nIt does lots of things.",
        ),
    )
    results1 = validate_docstring(class_bad_method, property_ignore=["EX01", "SA01"])
    assert results1["error_count"] == 1, results1["errors"]

    results2 = validate_docstring(class_bad_method)
    assert results2["error_count"] == 2
    assert "GL08" in results2["errors"]
    del results2["errors"]["GL08"]
    assert all(
        msg.startswith("Class.prop") for msg in chain(*results2["errors"].values())
    )


def tagged_f():
    # doc-ignore: GL08
    # doc-ignore: RT03 RT05
    return


class TaggedClass:
    # doc-ignore: GL08,GL02    RT01  ,  EX01

    def prop(self):
        """Bye"""
        # doc-ignore: YD01,SS06
        ...

    def f(self):
        """Hello"""
        # doc-ignore: SA04 not-a-code EX01, SA04
        ...


@pytest.mark.parametrize(
    "obj,expected_codes",
    [
        (tagged_f, {"GL08", "RT03", "RT05"}),
        (TaggedClass, {"GL08", "GL02", "RT01", "EX01"}),
        (TaggedClass.f, {"EX01", "SA04"}),
        (TaggedClass.prop, {"YD01", "SS06"}),
    ],
)
def test_tag_parsing(obj, expected_codes):
    assert _get_numpy_tags(obj) == expected_codes


class A:
    # doc-ignore: GL08

    def f(self):
        """Hello"""
        # doc-ignore: SA04 not-a-code ES01 SS03, EX01
        ...


class B:
    # doc-ignore: GL08

    def f(self):
        """Hello"""
        # doc-ignore: SA04 not-a-code NOQA
        ...


@pytest.mark.parametrize("obj, num_err", [(tagged_f, 1), (A, 4), (B, 4)])
def test_validates_respects_comments(obj, num_err):
    assert validate_docstring(obj)["error_count"] == 0, validate_docstring(obj)
    assert (
        validate_docstring(obj, ignore_via_comments_allowed=False)["error_count"]
        == num_err
    )


class Parent:
    def f(self):
        ...


class Child(Parent):
    # doc-ignore: GL08
    #
    # method f should not be scanned
    ...


class AProtocol(Protocol):
    def meth(self):
        ...


class ATypedDict(TypedDict):
    x: int


@pytest.mark.parametrize("obj", [Child, AProtocol, ATypedDict])
def test_no_scan(obj):
    assert validate_docstring(obj)["error_count"] == 0, validate_docstring(obj)



# Copyright 2024, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT

# ----
# Not exposing provider protocols publicly yet on the grounds that writing provider
# implementers will likely affect protocol design.
# ----



# Copyright 2024, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT

import re
from enum import Enum, Flag
from functools import partial
from typing import Any, Collection, NamedTuple, Tuple, Union

import hypothesis.strategies as st
import pytest
from hypothesis import assume, given, settings

from maite.errors import InvalidArgument
from maite.utils.validation import (
    chain_validators,
    check_domain,
    check_one_of,
    check_type,
)


def everything_except(excluded_types):
    return (
        st.from_type(type)
        .flatmap(st.from_type)
        .filter(lambda x: not isinstance(x, excluded_types))
    )


any_types = st.from_type(type)


@settings(max_examples=10)
@given(
    name=st.sampled_from(["name_a", "name_b"]),
    target_type=st.shared(any_types, key="target_type"),
    arg=st.shared(any_types, key="target_type").flatmap(everything_except),
)
def test_check_type_catches_bad_type(name, target_type, arg):
    with pytest.raises(InvalidArgument):
        check_type(name, arg=arg, type_=target_type)


@settings(deadline=None)
@given(
    target_type=st.shared(any_types, key="target_type"),
    arg=st.shared(any_types, key="target_type").flatmap(st.from_type),
)
def test_check_type_passes_good_type(target_type, arg):
    check_type("dummy", arg=arg, type_=target_type)


@given(...)
def test_check_multiple_types(arg: Union[str, int, None]):
    out = check_type("dummy", arg, type_=(str, int), optional=True)
    assert out == arg


def test_no_bounds():
    with pytest.raises(AssertionError):
        check_domain("arg", 1)  # pyright: ignore


@pytest.mark.parametrize(
    "kwargs",
    [
        pytest.param(
            {"arg": 1, "lower": 1, "upper": 1, "incl_low": False, "incl_up": False},
            marks=pytest.mark.xfail(raises=AssertionError, strict=True),
            id="1 < ... < 1",
        ),
        pytest.param(
            {"arg": 1, "lower": 1, "upper": 1, "incl_low": True, "incl_up": False},
            marks=pytest.mark.xfail(raises=AssertionError, strict=True),
            id="1 <= ... < 1",
        ),
        pytest.param(
            {"arg": 1, "lower": 1, "upper": 1, "incl_low": True, "incl_up": True},
            id="1 <= ... <= 1",
        ),
        pytest.param(
            {"arg": 1, "lower": 1, "upper": 1, "incl_low": False, "incl_up": True},
            marks=pytest.mark.xfail(raises=AssertionError, strict=True),
            id="1 < ... <= 1",
        ),
        pytest.param(
            {"arg": 1, "lower": 2, "upper": 1, "incl_low": False, "incl_up": False},
            marks=pytest.mark.xfail(raises=AssertionError, strict=True),
            id="2 < ... < 1",
        ),
        pytest.param(
            {"arg": 1, "lower": 2, "upper": 1, "incl_low": True, "incl_up": False},
            marks=pytest.mark.xfail(raises=AssertionError, strict=True),
            id="2 <= ... < 1",
        ),
        pytest.param(
            {"arg": 1, "lower": 2, "upper": 1, "incl_low": False, "incl_up": True},
            marks=pytest.mark.xfail(raises=AssertionError, strict=True),
            id="2 < ... <= 1",
        ),
        pytest.param(
            {"arg": 1, "lower": 2, "upper": 1, "incl_low": True, "incl_up": True},
            marks=pytest.mark.xfail(raises=AssertionError, strict=True),
            id="2 <= ... <= 1",
        ),
    ],
)
def test_min_max_ordering(kwargs):
    check_domain("dummy", **kwargs)


@pytest.mark.parametrize(
    "kwargs",
    [
        pytest.param(
            {"arg": 1, "lower": 1, "incl_low": False},
            marks=pytest.mark.xfail(raises=InvalidArgument, strict=True),
            id="lower:1 < arg:1",
        ),
        pytest.param({"arg": 1, "lower": 1, "incl_low": True}, id="lower:1 <= arg:1"),
        pytest.param(
            {"arg": 1, "upper": 1, "incl_up": False},
            marks=pytest.mark.xfail(raises=InvalidArgument, strict=True),
            id="arg:1 < upper:1",
        ),
        pytest.param({"arg": 1, "upper": 1, "incl_up": True}, id="arg:1 <= upper:1"),
        pytest.param(
            {"arg": 1, "lower": 1, "upper": 1, "incl_low": True, "incl_up": True},
            id="lower:1 <= arg:1 <= upper:1",
        ),
    ],
)
def test_bad_inequality(kwargs):
    check_domain("dummy", **kwargs)


@given(
    lower=(st.none() | st.floats(allow_nan=False, min_value=-1e6, max_value=1e6)),
    upper=(st.none() | st.floats(allow_nan=False, min_value=-1e6, max_value=1e6)),
    data=st.data(),
    incl_up=st.booleans(),
    incl_low=st.booleans(),
)
def test_valid_inequalities(
    lower, upper, data: st.DataObject, incl_up: bool, incl_low: bool
):
    if lower is None:
        incl_low = True
    if upper is None:
        incl_up = True

    if lower is None and upper is None:
        assume(False)
        assert False

    if lower is not None and upper is not None:
        lower, upper = (upper, lower) if upper < lower else (lower, upper)

    if incl_low is False or incl_up is False and lower == upper:
        assume(False)
        assert False
    if lower is None and upper is None:
        assume(False)
        assert False

    arg = (
        data.draw(
            st.floats(
                min_value=lower,
                max_value=upper,
                exclude_max=not incl_up,
                exclude_min=not incl_low,
            ),
            label="arg",
        )
        if lower != upper
        else lower
    )

    check_domain(
        "dummy",
        arg=arg,
        lower=lower,
        upper=upper,
        incl_up=incl_up,
        incl_low=incl_low,
    )


class CheckOneOfInputs(NamedTuple):
    arg: Any
    collection: Collection = []
    vals: Tuple[Any, ...] = ()
    requires_identity: bool = False
    name: str = "foo"


class AClass:
    pass


class BClass:
    pass


FooEnum = Enum("Foo", ["a", "b"])
BarEnum = Enum("Bar", ["c", "d"])
FlagEnum = Flag("Flag", ["e", "f"])

collections = st.lists(st.sampled_from([True, BClass(), BClass]), unique=True)


@given(
    arg=st.sampled_from([None, False, AClass(), AClass, FooEnum, FooEnum.a])
    | st.integers(-5, 0),
    collection=collections | st.just(BarEnum) | st.just(FlagEnum),
    vals=collections,
    requires_identity=st.booleans(),
)
def test_check_one_of_catches_bad_inputs(arg, collection, vals, requires_identity):
    if not collection and not vals:
        assume(False)

    with pytest.raises(InvalidArgument):
        check_one_of("foo", arg, collection, *vals, requires_identity=requires_identity)


@given(...)
def test_check_one_of_supports_enum(arg: Union[FooEnum, FlagEnum]):
    assert check_one_of("foo", arg, type(arg)) is arg


@given(
    arg=st.sampled_from([None, False, AClass(), AClass, FooEnum, FooEnum.a])
    | st.integers(-5, 0),
    collection=collections,
    vals=collections,
    requires_identity=st.booleans(),
)
def test_check_one_of_passes(
    arg, collection: list, vals: list, requires_identity: bool
):
    collection.append(arg)
    assert (
        check_one_of("foo", arg, collection, *vals, requires_identity=requires_identity)
        is arg
    )
    assert (
        check_one_of("foo", arg, vals, *collection, requires_identity=requires_identity)
        is arg
    )


def test_check_one_of_raises_unsatisfiable():
    with pytest.raises(AssertionError):
        check_one_of("foo", 1, [])


is_int = partial(check_type, type_=int)
is_pos = partial(check_domain, lower=0)
check_pos_int = chain_validators(is_int, is_pos)


@given(st.integers(0, 10))
def test_chain_validators_pass_through(x: int):
    assert x == check_pos_int("arg", x)


@pytest.mark.parametrize(
    "expr, msg",
    [
        (
            lambda: check_domain("arg", 1, lower=2, incl_low=False),
            r"`arg` must satisfy `2 < arg`.  Got: `1`.",
        ),
        (
            lambda: check_domain("arg", 1, lower=2, incl_low=True),
            r"`arg` must satisfy `2 <= arg`.  Got: `1`.",
        ),
        (
            lambda: check_domain("arg", 1, lower=2, incl_low=False, lower_name="low"),
            r"`arg` must satisfy `low=2 < arg`.  Got: `1`.",
        ),
        (
            lambda: check_domain("arg", 1, lower=2, incl_low=True, lower_name="low"),
            r"`arg` must satisfy `low=2 <= arg`.  Got: `1`.",
        ),
        (
            lambda: check_domain("arg", 1, upper=0, incl_up=False),
            r"`arg` must satisfy `arg < 0`.  Got: `1`.",
        ),
        (
            lambda: check_domain("arg", 1, upper=0, incl_up=True),
            r"`arg` must satisfy `arg <= 0`.  Got: `1`.",
        ),
        (
            lambda: check_domain("arg", 1, upper=0, incl_up=False, upper_name="hi"),
            r"`arg` must satisfy `arg < hi=0`.  Got: `1`.",
        ),
        (
            lambda: check_domain("arg", 1, upper=0, incl_up=True, upper_name="hi"),
            r"`arg` must satisfy `arg <= hi=0`.  Got: `1`.",
        ),
        (
            lambda: check_type("arg", 1, str),
            r"Expected `arg` to be of type `str`. Got `1` (type: `int`).",
        ),
        (
            lambda: check_type("arg", 1, (str, bool)),
            r"Expected `arg` to be of types: `str`, `bool`. Got `1` (type: `int`).",
        ),
        (
            lambda: check_type("arg", 1, (str, bool), optional=True),
            r"Expected `arg` to be `None` or of types: `str`, `bool`. Got `1` (type: `int`).",
        ),
        (
            lambda: check_one_of("foo", 1, [], requires_identity=True),
            r"`collections` and `args` are both empty.",
        ),
        (
            lambda: check_one_of("foo", 1, [True], requires_identity=True),
            r"Expected `foo` to be: True. Got `1`.",
        ),
        (
            lambda: check_one_of("foo", 1, [], True, requires_identity=True),
            r"Expected `foo` to be: True. Got `1`.",
        ),
        (
            lambda: check_one_of("bar", 1, [True, False], requires_identity=True),
            r"Expected `bar` to be one of: False, True. Got `1`.",
        ),
        (
            lambda: check_one_of("bar", 1, [True, False], 2, requires_identity=True),
            r"Expected `bar` to be one of: 2, False, True. Got `1`.",
        ),
        (
            lambda: check_one_of("bar", 1, [], True, False, 2, requires_identity=True),
            r"Expected `bar` to be one of: 2, False, True. Got `1`.",
        ),
        (
            # error message should remove redundant names
            lambda: check_one_of(
                "bar", 1, [True], True, False, 2, requires_identity=True
            ),
            r"Expected `bar` to be one of: 2, False, True. Got `1`.",
        ),
        (
            lambda: check_one_of("bar", [1], Enum("Foo", ["a", "b"])),
            r"Expected `bar` to be one of: Foo.a, Foo.b. Got `[1]`.",
        ),
        (
            lambda: check_pos_int("foo", ["a"]),
            r"Expected `foo` to be of type `int`. Got `['a']` (type: `list`).",
        ),
        (
            lambda: check_pos_int("foo", -1),
            r"`foo` must satisfy `0 <= foo`.  Got: `-1`.",
        ),
    ],
)
def test_error_msg(expr, msg: str):
    with pytest.raises(Exception, match=re.escape(msg)):
        expr()



# Copyright 2024, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT

# import component generics from generic.py and specialize them for object detection
# domain

from __future__ import annotations

from typing import Any, Callable, Iterable, Protocol, Sequence, Tuple, runtime_checkable

from typing_extensions import Dict, TypeAlias

from maite._internals.protocols import generic as gen
from maite.protocols import ArrayLike

# We *could* make ArrayLike generic and rely on the subscripts for ArrayLike type
# annotations to hint to the user the appropriate shape. No runtime safety would
# be added by this approach because type subscripts are effectively invisible
# at runtime. No additional static type-checking would occur either, but the
# user would get useful type hints when cursoring over required inputs/targets
# This would also require python 3.11 unless a `from __future__` import were made
# available for earlier versions (which are not available now.)


@runtime_checkable
class ObjectDetectionTarget(Protocol):
    """
    An object-detection target protocol.

    This class is used to encode both predictions and ground-truth labels in the object
    detection problem.

    Implementers must populate the following attributes:

    Attributes
    ----------
    boxes : ArrayLike
        An array representing object detection boxes in a single image with x0, y0, x1, y1
        format and shape `(N_DETECTIONS, 4)`

    labels : ArrayLike
        An array representing the integer labels associated with each detection box of shape
        `(N_DETECTIONS,)`

    scores: ArrayLike
        An array representing the scores associated with each box (of shape `(N_DETECTIONS,)`)
    """

    @property
    def boxes(
        self,
    ) -> ArrayLike:  # shape (N, 4), format X0, Y0, X1, Y1
        ...

    @property
    def labels(self) -> ArrayLike:  # label for each box, shape (N,)
        ...

    @property
    def scores(self) -> ArrayLike:  # shape (N,)
        ...


# TODO: remove typeAlias statements for more user readability (or figure out how to resolve TypeAliases
#       to their targets for end-user.) Knowing a dataset returns a tuple of "InputType, TargetType, DatumMetadataType"
#       isn't helpful to implementers, however the aliasing *is* helpful to developers.
#
#       Perhaps the functionality I want is named TypeVars for generic, so developers can understand that
#       e.g. generic.Dataset typevars are 'InputType', 'TargetType', and 'MetaDataType' and their values in
#       concrete Dataset classes (like object_detection.Dataset) are ArrayLike, ObjectDetectionTarget, Dict[str,Any]
#       so users can see an expected return type of Tuple[ArrayLike, ObjectDetectionTarget, Dict[str,Any]]

InputType: TypeAlias = ArrayLike  # shape (C, H, W)
TargetType: TypeAlias = ObjectDetectionTarget
DatumMetadataType: TypeAlias = Dict[str, Any]

InputBatchType: TypeAlias = Sequence[
    ArrayLike
]  # sequence of N ArrayLikes of shape (C, H, W)
TargetBatchType: TypeAlias = Sequence[TargetType]  # sequence of N TargetType instances
DatumMetadataBatchType: TypeAlias = Sequence[DatumMetadataType]

Datum: TypeAlias = Tuple[InputType, TargetType, DatumMetadataType]
DatumBatch: TypeAlias = Tuple[InputBatchType, TargetBatchType, DatumMetadataBatchType]

CollateFn: TypeAlias = Callable[
    [Iterable[Datum]],
    DatumBatch,
]


class Dataset(gen.Dataset[InputType, TargetType, DatumMetadataType], Protocol):
    """
    A dataset protocol for object detection ML subproblem providing datum-level
    data access.

    Implementers must provide index lookup (via `__getitem__(ind: int)` method) and
    support `len` (via `__len__()` method). Data elements looked up this way correspond to
    individual examples (as opposed to batches).

    Indexing into or iterating over the an object detection dataset returns a `Tuple` of
    types `ArrayLike`, `ObjectDetectionTarget`, and `Dict[str,Any]`. These correspond to
    the model input type, model target type, and datum-level metadata, respectively.


    Methods
    -------

    __getitem__(ind: int) -> Tuple[ArrayLike, ObjectDetectionTarget, Dict[str, Any]]
        Provide mapping-style access to dataset elements. Returned tuple elements
        correspond to model input type, model target type, and datum-specific metadata,
        respectively.

    __len__() -> int
        Return the number of data elements in the dataset.

    """

    ...


class DataLoader(
    gen.DataLoader[
        InputBatchType,
        TargetBatchType,
        DatumMetadataBatchType,
    ],
    Protocol,
):
    """
    A dataloader protocol for the object detection ML subproblem providing
    batch-level data access.

    Implementers must provide an iterable object (returning an iterator via the
    `__iter__` method) that yields tuples containing batches of data. These tuples
    contain types `Sequence[ArrayLike]` (elements of shape `(C, H, W)`),
    `Sequence[ObjectDetectionTarget]`, and `Sequence[Dict[str, Any]]`,
    which correspond to model input batch, model target batch, and a datum metadata batch.

    Note: Unlike Dataset, this protocol does not require indexing support, only iterating.


    Methods
    -------

    __iter__ -> Iterator[tuple[Sequence[ArrayLike], Sequence[ObjectDetectionTarget], Sequence[Dict[str, Any]]]]
        Return an iterator over batches of data, where each batch contains a tuple of
        of model input batch (as `Sequence[ArrayLike]`), model target batch (as
        `Sequence[ObjectDetectionTarget]`), and batched datum-level metadata
        (as `Sequence[Dict[str,Any]]`), respectively.

    """

    ...


class Model(gen.Model[InputBatchType, TargetBatchType], Protocol):
    """
    A model protocol for the object detection ML subproblem.

    Implementers must provide a `__call__` method that operates on a batch of model inputs
    (as `Sequence[ArrayLike]`s) and returns a batch of model targets (as
    `Sequence[ObjectDetectionTarget]`)

    Methods
    -------

    __call__(input_batch: Sequence[ArrayLike]) -> Sequence[ObjectDetectionTarget]
        Make a model prediction for inputs in input batch. Elements of input batch
        are expected in the shape `(C, H, W)`.
    """

    ...


class Metric(gen.Metric[TargetBatchType], Protocol):
    """
    A metric protocol for the object detection ML subproblem.

     A metric in this sense is expected to measure the level of agreement between model
     predictions and ground-truth labels.

     Methods
     -------

     update(preds: Sequence[ObjectDetectionTarget], targets: Sequence[ObjectDetectionTarget]) -> None
         Add predictions and targets to metric's cache for later calculation.

     compute() -> Dict[str, Any]
         Compute metric value(s) for currently cached predictions and targets, returned as
         a dictionary.

     reset() -> None
         Clear contents of current metric's cache of predictions and targets.
    """

    ...


class Augmentation(
    gen.Augmentation[
        InputBatchType,
        TargetBatchType,
        DatumMetadataBatchType,
        InputBatchType,
        TargetBatchType,
        DatumMetadataBatchType,
    ],
    Protocol,
):
    """
    An augmentation protocol for the object detection subproblem.

    An augmentation is expected to take a batch of data and return a modified version of
    that batch. Implementers must provide a single method that takes and returns a
    labeled data batch, where a labeled data batch is represented by a tuple of types
    `Sequence[ArrayLike]`, `Sequence[ObjectDetectionTarget]`, and `Sequence[Dict[str,Any]]`.
    These correspond to the model input batch type, model target batch type, and datum-level
    metadata batch type, respectively.

    Methods
    -------

    __call__(datum: Tuple[Sequence[ArrayLike], Sequence[ObjectDetectionTarget], Sequence[dict[str, Any]]]) ->\
          Tuple[Sequence[ArrayLike], Sequence[ObjectDetectionTarget], Sequence[dict[str, Any]]]
        Return a modified version of original data batch. A data batch is represented
        by a tuple of model input batch (as `Sequence ArrayLike` with elements of shape
        `(C, H, W)`), model target batch (as `Sequence[ObjectDetectionTarget]`), and
        batch metadata (as `Sequence[Dict[str,Any]]`), respectively.
    """

    ...



# Copyright 2024, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT

"""
Utilities for validating documentation strings for maite projects.
"""

from maite._internals.testing.docs import (
    NumpyDocErrorCode,
    NumPyDocResults,
    validate_docstring,
)

__all__ = ["validate_docstring", "NumpyDocErrorCode", "NumPyDocResults"]



# Copyright 2024, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT

# flake8: noqa
import re
from collections import defaultdict
from inspect import getsource, isclass
from itertools import chain, zip_longest
from typing import (
    Any,
    Callable,
    Collection,
    Dict,
    List,
    Literal,
    Optional,
    Protocol,
    Set,
    Tuple,
    Union,
    cast,
    overload,
)

from typing_extensions import NotRequired, TypeAlias, TypedDict

from maite._internals.validation import check_type
from maite.errors import InvalidArgument

from ..utils import is_typed_dict

NumpyDocErrorCode: TypeAlias = Literal[
    "GL01",
    "GL02",
    "GL03",
    "GL05",
    "GL06",
    "GL07",
    "GL08",
    "GL09",
    "GL10",
    "SS01",
    "SS02",
    "SS03",
    "SS04",
    "SS05",
    "SS06",
    "ES01",
    "PR01",
    "PR02",
    "PR03",
    "PR04",
    "PR05",
    "PR06",
    "PR07",
    "PR08",
    "PR09",
    "PR10",
    "RT01",
    "RT02",
    "RT03",
    "RT04",
    "RT05",
    "YD01",
    "SA01",
    "SA02",
    "SA03",
    "SA04",
    "EX01",
    "NOQA",
]

ERRORCODES: Set[NumpyDocErrorCode] = set(NumpyDocErrorCode.__args__)  # type: ignore


class _C:
    ...


AUTO_INIT_DOC = _C.__init__.__doc__


class _NumpyDocValidate(TypedDict):
    # doc-ignore: NOQA
    type: Literal["function", "type"]
    docstring: str
    deprecated: bool
    file: str
    file_line: int
    errors: List[Tuple[NumpyDocErrorCode, str]]


class NumPyDocResults(TypedDict):
    # doc-ignore: NOQA
    error_count: int
    errors: Dict[NumpyDocErrorCode, List[str]]
    file: str
    file_line: int
    ignored_errors: NotRequired[Dict[NumpyDocErrorCode, List[str]]]


class NumPyDocResultsWithIgnored(NumPyDocResults):
    # doc-ignore: NOQA
    ignored_errors: Dict[NumpyDocErrorCode, List[str]]


doc_ignore_re = re.compile(r"#\s?doc-ignore:(.*)")
_comma_or_whitespace = re.compile(r"[,\s+]")


def _get_numpy_tags(obj: Any) -> Set[NumpyDocErrorCode]:
    """Searches source code for # doc-ignore: <list of error codes>."""
    try:
        src = getsource(obj)
    except TypeError:  # pragma: no cover
        # handles classes defined in `__main__`
        return set()
    if isclass(obj):
        src = src[: src.find("def ")]
    joined_tags = ",".join([x.strip() for x in re.findall(doc_ignore_re, src)])
    joined_tags = set(x.strip() for x in re.split(_comma_or_whitespace, joined_tags))
    return ERRORCODES & joined_tags


@overload
def validate_docstring(
    obj: Any,
    ignore: Collection[NumpyDocErrorCode] = ...,
    method_ignore: Optional[Collection[NumpyDocErrorCode]] = ...,
    property_ignore: Optional[Collection[NumpyDocErrorCode]] = ...,
    include_ignored_errors: Literal[True] = ...,
    ignore_via_comments_allowed: bool = ...,
) -> NumPyDocResultsWithIgnored:
    ...


@overload
def validate_docstring(
    obj: Any,
    ignore: Collection[NumpyDocErrorCode] = ...,
    method_ignore: Optional[Collection[NumpyDocErrorCode]] = ...,
    property_ignore: Optional[Collection[NumpyDocErrorCode]] = ...,
    include_ignored_errors: bool = ...,
    ignore_via_comments_allowed: bool = ...,
) -> NumPyDocResults:
    ...


def validate_docstring(
    obj: Any,
    ignore: Collection[NumpyDocErrorCode] = ("SA01",),
    method_ignore: Optional[Collection[NumpyDocErrorCode]] = None,
    property_ignore: Optional[Collection[NumpyDocErrorCode]] = None,
    include_ignored_errors: bool = False,
    ignore_via_comments_allowed: bool = True,
) -> Union[NumPyDocResults, NumPyDocResultsWithIgnored]:
    # fmt: off
    """
    Validate an object's docstring against the NumPy docstring standard [1]_.

    The body of a function or class can include a comment of the format:: 

       # doc-ignore: <CODE1> <CODE2> [...]
    
    where each <CODEN> is any error code listed in the Notes section. This will cause
    `validate_docstring` to ignore said error code during its analysis.

    When `obj` is a class object, the `__doc__` attributes of `obj`, `obj.__init__` and
    of all of the public methods and properties of `obj` will be validated. The
    doc strings of `doc` and `doc.__init__` are considered in conjunction with one
    another so that users need not specify redundant documentation.

    This function require that `numpydoc` is installed.

    Parameters
    ----------
    obj : Any
        A function, class, module, or method whose docstring will be validated.
        If `obj` is a class, its public methods and properties will be traversed and
        their doc strings will be included in the validation process.

    ignore : Collection[ErrorCode], optional (default=['SA01'])
        One or more error codes to be excluded from the reported errors. See the
        Notes section for a list of error codes. NOQA can be specified to ignore
        *all* errors.

    method_ignore : Optional[Collection[ErrorCode]]
        For method docstring. One or more error codes to be excluded from the reported
        errors. If not specified, defers to the codes specified in `ignore`.

    property_ignore : Optional[Collection[ErrorCode]]
        For property doc strings. One or more error codes to be excluded from the
        reported errors. If not specified, defers to the codes specified in `ignore`.

    include_ignored_errors : bool, optional (default=False)
        If `True`, include the errors that were ignored during the validation.

    ignore_via_comments_allowed : bool, optional (default=True)
        If `True` then the source code of `obj` will be parsed for comments of the form
        # doc-ignore: <CODE1> <CODE2> [...] to extract additional error codes that 
        will be ignored during the validation process. Class properties are not 
        supported.

    Returns
    -------
    NumPyDocResults
        A dictionary with the following fields.
            - error_count : int
            - errors : Dict[ErrorCode, List[str]]
            - ignored_errors : NotEquired[Dict[ErrorCode, List[str]]]
            - file : str
            - file_line : int

    Notes
    -----
    The following are the error codes that can be returned by this validation function:
    - NOQA: Can be specified to ignore all error codes.
    - GL01: Docstring text (summary) should start in the line immediately after the opening quotes.
    - GL02: Closing quotes should be placed in the line after the last text in the docstring.
    - GL03: Double line break found.
    - GL05: Tabs found at the start of line.
    - GL06: Found unknown section.
    - GL07: Sections are in the wrong order.
    - GL08: The object does not have a docstring.
    - GL09: Deprecation warning should precede extended summary.
    - GL10: reST directives {directives} must be followed by two colons.
    - SS01: No summary found.
    - SS02: Summary does not start with a capital letter.
    - SS03: Summary does not end with a period.
    - SS04: Summary contains heading whitespaces.
    - SS05: Summary must start with infinitive verb, not third person.
    - SS06: Summary should fit in a single line.
    - ES01: No extended summary found.
    - PR01: Signature parameter not documented.
    - PR02: Unknown parameters included in Parameters.
    - PR03: Wrong parameters order.
    - PR04: Parameter has no type
    - PR05: Parameter type should not finish with ".".
    - PR06: Parameter type should be changed.
    - PR07: Parameter has no description.
    - PR08: Parameter description should start with a capital letter.
    - PR09: Parameter description should finish with ".".
    - PR10: Parameter requires a space before the colon.
    - RT01: No Returns section found
    - RT02: The first line of the Returns section should contain only the type.
    - RT03: Return value has no description.
    - RT04: Return value description should start with a capital letter.
    - RT05: Return value description should finish with ".".
    - YD01: No Yields section found.
    - SA01: See Also section not found.
    - SA02: Missing period at end of description for See Also reference.
    - SA03: Description should be capitalized for See Also reference.
    - SA04: Missing description for See Also reference.
    - EX01: No examples section found.

    References
    ----------
    .. [1] https://numpydoc.readthedocs.io/en/latest/format.html

    Examples
    --------
    >>> from maite.testing.docs import validate_docstring
    >>> from maite.testing.documentation.documentation_dependencies import person
    
    Let's ignore the need for an Extended Summary and a See Also section.

    >>> validate_docstring(person, ignore=('ES01', 'SA01'), include_ignored_errors=True)
    {'error_count': 0, 'errors': {}, ...'ignored_errors': {'ES01': ['No extended summary found'], 'SA01': ['See Also section not found']}...
    
    Using comments to skip validation.

    >>> def f():
    ...     # doc-ignore: GL08
    ...     return
    >>> validate_docstring(f)
    {'error_count': 0, 'errors': {}, 'file': None, 'file_line': 1}
    >>> validate_docstring(f, ignore_via_comments_allowed=False)
    {'error_count': 1, 'errors': {'GL08': ['The object does not have a docstring']}, 'file': None, 'file_line': 1}
    """
    # fmt: on
    try:
        from numpydoc.docscrape import ClassDoc
        from numpydoc.validate import get_doc_object, validate
    except ImportError:
        raise ImportError("`numpydoc` must be installed in order to use this function.")

    check_type("ignore_via_comments_allowed", ignore_via_comments_allowed, bool)

    ignore = set(ignore)

    if method_ignore is None:
        method_ignore = ignore

    if property_ignore is None:
        property_ignore = ignore

    method_ignore = set(method_ignore)
    property_ignore = set(property_ignore)

    get_tags = (
        _get_numpy_tags
        if ignore_via_comments_allowed
        else lambda _: cast(Set[NumpyDocErrorCode], set())
    )

    for _name, _codes in [
        ("ignore", ignore),
        ("method_ignore", method_ignore),
        ("property_ignore", property_ignore),
    ]:
        if not _codes <= ERRORCODES:
            unknown = ", ".join(sorted(_codes - ERRORCODES))
            raise InvalidArgument(
                f"`{_name}` contains the following elements that are not valid error "
                f"code(s): {unknown}"
            )
    validate = cast(Callable[[Any], _NumpyDocValidate], validate)

    errors: Dict[NumpyDocErrorCode, List[str]] = defaultdict(list)
    ignored_errors: Dict[NumpyDocErrorCode, List[str]] = defaultdict(list)

    def update_errors(
        new_errors: List[Tuple[NumpyDocErrorCode, str]],
        ignore: Set[NumpyDocErrorCode],
        prefix: str = "",
    ) -> None:
        if "NOQA" in ignore:
            return

        for err_code, err_msg in new_errors:
            if err_code not in ignore:
                errors[err_code].append(prefix + err_msg)
            else:
                ignored_errors[err_code].append(prefix + err_msg)

    if is_typed_dict(obj):
        ignore.add("NOQA")
    elif isinstance(obj, type) and issubclass(obj, Protocol):
        ignore.add("NOQA")

    doc_obj = get_doc_object(obj)

    results = validate(doc_obj)
    results_codes = set(c for c, _ in results["errors"])

    update_errors(results["errors"], ignore=get_tags(obj) | ignore)

    # We want to avoid forcing redundant sections between the class-level
    # docstring and the docstring for the class' __init__. Here we ignore
    # an error in one section if the error is not present in the other section.
    #
    # E.g., the class-level doc can be missing the Parameters section as long
    # as the __init__ is not also missing that section.
    if isinstance(doc_obj, ClassDoc):
        init_results = validate(get_doc_object(obj.__init__))
        if obj.__init__.__doc__ == AUTO_INIT_DOC:
            init_results["errors"].append(
                ("GL08", "The object does not have a docstring")
            )

        init_codes = set(c for c, _ in init_results["errors"])

        if "GL08" not in init_codes:
            update_errors(
                init_results["errors"],
                prefix=f"{obj.__name__}.__init__: ",
                ignore=get_tags(obj.__init__) | ignore,
            )
            # Ignore 'missing section' errors unless the error occurs in both the
            # class docstring and in the __init__ docstring
            resolved: List[NumpyDocErrorCode] = [
                code
                for code in errors
                if (code.endswith("01") or code == "GL08")
                and ((code not in init_codes) ^ (code not in results_codes))
            ]

            for item in resolved:
                errors.pop(item)

        del init_results

        for _ignore, name in chain(
            zip_longest([], doc_obj.properties, fillvalue=property_ignore),
            zip_longest([], doc_obj.methods, fillvalue=method_ignore),
        ):
            if isinstance(obj, type) and issubclass(obj, Protocol):
                # protocol attributes don't need docstrings
                break

            if name not in obj.__dict__:
                # don't scan inherited methods/attrs
                continue

            assert isinstance(name, str)
            _ignore = cast(Set[NumpyDocErrorCode], _ignore)
            _member = getattr(obj, name)
            attr_results = validate(get_doc_object(_member))
            prefix = f"{obj.__name__}.{name}: "
            update_errors(
                attr_results["errors"],
                prefix=prefix,
                ignore=get_tags(_member) | _ignore,
            )

    out = NumPyDocResults(
        error_count=sum((len(x) for x in errors.values())),
        errors=dict(errors),
        file=results["file"],
        file_line=results["file_line"],
    )
    if include_ignored_errors:
        out["ignored_errors"] = dict(ignored_errors)
    return out



import copy
from dataclasses import dataclass
from typing import Any, Dict, Iterator, List, Sequence, Tuple

import numpy as np

from maite.protocols.object_detection import (
    DatumMetadataBatchType,
    InputBatchType,
    ObjectDetectionTarget,
    TargetBatchType,
)

# test lightweight implementations
#
# mock up a set of components such that:
#
# InputType = np.array of shape [C, H, W]
# TargetType = ObjectDetectionTarget
# DatumMetadataType = Dict[str, Any]
#
# InputBatchType = np.array of shape [N, C, H, W]
# TargetBatchType = Sequence[ObjectDetectionTarget]
# DatumMetadataBatchType = list[Dict[str, Any]]


N_CLASSES = 5  # how many classes
N_DATAPOINTS = 10  # datapoints in dataset
C = 3  # number of color channels
H = 5  # img height
W = 4  # img width
OBJ_PER_IMG = 3  # number of (planned) objects per image
OBJ_PER_MODEL_PRED = 2


@dataclass
class ObjectDetectionTargetImpl:
    boxes: np.ndarray  # shape: (N_DETECTIONS, 4), X0,Y0,X1,Y1
    scores: np.ndarray  # shape: (N_DETECTIONS,)
    labels: np.ndarray  # shape: (N_DETECTIONS,)

    @classmethod
    def from_protocol(
        cls, prot_obj: ObjectDetectionTarget
    ) -> "ObjectDetectionTargetImpl":
        return cls(
            boxes=np.array(prot_obj.boxes),
            scores=np.array(prot_obj.scores),
            labels=np.array(prot_obj.labels),
        )


class DatasetImpl:
    def __init__(self):
        self._data: np.ndarray = np.random.rand(N_DATAPOINTS, C, H, W)

        self._targets: List[ObjectDetectionTargetImpl] = [
            ObjectDetectionTargetImpl(
                boxes=np.array(
                    [
                        [
                            i / (OBJ_PER_IMG) * (W - 1),
                            i / (OBJ_PER_IMG) * (H - 1),
                            i / (OBJ_PER_IMG) * (W - 1) + 1,
                            i / (OBJ_PER_IMG) * (H - 1) + 1,
                        ]
                        for i in range(OBJ_PER_IMG)
                    ]
                ),
                scores=np.linspace(0, 1, OBJ_PER_IMG),
                labels=np.array([i % N_CLASSES for i in range(OBJ_PER_IMG)]),
            )
            for _ in range(N_DATAPOINTS)
        ]

        self._data_metadata: Sequence[Dict[str, Any]] = [
            {"some_metadata": i} for i in range(self._data.shape[0])
        ]

    def __len__(self) -> int:
        return self._data.shape[0]

    def __getitem__(
        self, ind: int
    ) -> Tuple[np.ndarray, ObjectDetectionTargetImpl, Dict[str, Any]]:
        return (self._data[ind], self._targets[ind], self._data_metadata[ind])


class DataLoaderImpl:
    def __init__(self, dataset: DatasetImpl, batch_size: int = 2):
        self._dataset = dataset
        self._batch_size = batch_size

    def __iter__(
        self,
    ) -> Iterator[
        Tuple[List[np.ndarray], List[ObjectDetectionTargetImpl], List[Dict[str, Any]]]
    ]:
        # calculate number of batches
        n_batches = len(self._dataset) // self._batch_size

        # handle last batch if dataset length not divisible by batch size
        if len(self._dataset) / self._batch_size > n_batches:
            n_batches += 1

        for i_batch in range(n_batches):
            # i_batch goes from 0->n_batches

            batch_data = [
                self._dataset[i]
                for i in range(
                    i_batch * self._batch_size,
                    np.min([(i_batch + 1) * self._batch_size, len(self._dataset)]),
                )
            ]
            batch_inputs: List[np.ndarray] = []
            batch_targets: List[ObjectDetectionTargetImpl] = []
            batch_mds: List[Dict[str, Any]] = []
            for batch_tup in batch_data:
                batch_inputs.append(np.array(batch_tup[0]))
                batch_targets.append(batch_tup[1])
                batch_mds.append(batch_tup[2])

            yield (batch_inputs, batch_targets, batch_mds)


class AugmentationImpl:
    def __init__(self):
        ...

    def __call__(
        self,
        __datum_batch: Tuple[InputBatchType, TargetBatchType, DatumMetadataBatchType],
    ) -> Tuple[
        List[np.ndarray], List[ObjectDetectionTargetImpl], Sequence[Dict[str, Any]]
    ]:
        input_batch_aug = copy.deepcopy([np.array(elem) for elem in __datum_batch[0]])
        target_batch_aug = copy.deepcopy(
            [ObjectDetectionTargetImpl.from_protocol(i) for i in __datum_batch[1]]
        )
        metadata_batch_aug = copy.deepcopy(__datum_batch[2])

        # -- manipulate input_batch, target_batch, and metadata_batch --

        # add new value to metadata_batch_aug
        for md in metadata_batch_aug:
            assert (
                "new_key" not in md.keys()
            ), "bad practice to write over keys in metadata"
            md["new_key"] = "new_val"

        # modify input batch
        for in_batch_elem in input_batch_aug:
            in_batch_elem += 1

        # modify target batch
        for target in target_batch_aug:
            target.scores = target.scores / 2

        return (input_batch_aug, target_batch_aug, metadata_batch_aug)


class ModelImpl:
    def __call__(
        self, __input_batch: InputBatchType
    ) -> Sequence[ObjectDetectionTargetImpl]:
        pass

        target_batch = [
            ObjectDetectionTargetImpl(
                boxes=np.array(
                    [
                        [
                            i / (OBJ_PER_MODEL_PRED) * (W - 1),
                            i / (OBJ_PER_MODEL_PRED) * (H - 1),
                            i / (OBJ_PER_MODEL_PRED) * (W - 1) + 1,
                            i / (OBJ_PER_MODEL_PRED) * (H - 1) + 1,
                        ]
                        for i in range(OBJ_PER_MODEL_PRED)
                    ]
                ),
                scores=np.linspace(0, 1, OBJ_PER_MODEL_PRED),
                labels=np.array([i % N_CLASSES for i in range(OBJ_PER_MODEL_PRED)]),
            )
            for _ in range(np.array(__input_batch).shape[0])
        ]

        return target_batch


class MetricImpl:
    def __init__(self):
        return None

    def reset(self) -> None:
        return None

    def update(
        self, __pred_batch: TargetBatchType, __target_batch: TargetBatchType
    ) -> None:
        return None

    def compute(self) -> Dict[str, Any]:
        return {"metric1": "val1", "metric2": "val2"}



# Copyright 2024, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT

"""
Functions for type-checking code using pyright.
"""
from maite._internals.testing.pyright import (
    Diagnostic,
    PyrightOutput,
    Summary,
    list_error_messages,
    pyright_analyze,
)

__all__ = [
    "Diagnostic",
    "PyrightOutput",
    "Summary",
    "list_error_messages",
    "pyright_analyze",
]



import copy
from typing import Any, Dict, Iterator, List, Sequence, Tuple

import numpy as np

from maite.protocols.image_classification import (
    DatumMetadataBatchType,
    InputBatchType,
    TargetBatchType,
)

# define lightweight component implementations
#
# pretend we have a set of components such that:
#
# InputType = np.array of shape [C, H, W]
# TargetType = np.array of shape [Cl]
# DatumMetadataType is an Dict[str, Any]
#
# InputBatchType = np.array of shape [N, C, H, W]
# TargetBatchType = np.array of shape [N, Cl]
# DatumMetadataBatchType = list[Dict[str, Any]]

N_CLASSES = 5  # how many classes
N_DATAPOINTS = 10  # datapoints in dataset
C = 3  # number of color channels
H = 5  # img height
W = 4  # img width


class DatasetImpl:
    def __init__(self):
        self._data = np.random.rand(N_DATAPOINTS, C, H, W)

        self._targets = np.zeros((N_DATAPOINTS, N_CLASSES))
        for data_index in range(self._targets.shape[0]):
            self._targets[data_index, data_index % N_CLASSES] = 1

        self._data_metadata = [{"some_metadata": i} for i in range(self._data.shape[0])]

    def __len__(self) -> int:
        return self._data.shape[0]

    def __getitem__(self, ind: int) -> Tuple[np.ndarray, np.ndarray, dict]:
        return (self._data[ind], self._targets[ind], self._data_metadata[ind])


class DataLoaderImpl:
    def __init__(self, dataset: DatasetImpl, batch_size=2):
        self._dataset = dataset
        self._batch_size = batch_size

    def __iter__(
        self,
    ) -> Iterator[
        Tuple[Sequence[np.ndarray], Sequence[np.ndarray], List[Dict[str, Any]]]
    ]:
        # calculate number of batches
        n_batches = len(self._dataset) // self._batch_size

        # handle last batch if dataset length not divisible by batch size
        if len(self._dataset) / self._batch_size > n_batches:
            n_batches += 1

        for i_batch in range(n_batches):
            # i_batch goes from 0->n_batches

            batch_data = [
                self._dataset[i]
                for i in range(
                    i_batch * self._batch_size,
                    np.min([(i_batch + 1) * self._batch_size, len(self._dataset)]),
                )
            ]
            batch_inputs = []
            batch_targets = []
            batch_mds = []
            for batch_tup in batch_data:
                batch_inputs.append(np.array(batch_tup[0]))
                batch_targets.append(np.array(batch_tup[1]))
                batch_mds.append(batch_tup[2])

            yield (batch_inputs, batch_targets, batch_mds)


class AugmentationImpl:
    def __init__(self):
        ...

    def __call__(
        self,
        __datum_batch: Tuple[InputBatchType, TargetBatchType, DatumMetadataBatchType],
    ) -> Tuple[List[np.ndarray], List[np.ndarray], Sequence[Dict[str, Any]]]:
        input_batch_aug = copy.deepcopy([np.array(elem) for elem in __datum_batch[0]])
        target_batch_aug = copy.deepcopy([np.array(elem) for elem in __datum_batch[1]])
        metadata_batch_aug = copy.deepcopy(__datum_batch[2])

        # -- manipulate input_batch, target_batch, and metadata_batch --

        # add new value to metadata_batch_aug
        for md in metadata_batch_aug:
            assert (
                "new_key" not in md.keys()
            ), "bad practice to write over keys in metadata"
            md["new_key"] = "new_val"

        # modify input batch
        for inp_batch_elem in input_batch_aug:
            inp_batch_elem += 1

        # modify target batch
        for tgt_batch_elem in target_batch_aug:
            tgt_batch_elem = np.mod(tgt_batch_elem + 1, N_CLASSES)

        return (input_batch_aug, target_batch_aug, metadata_batch_aug)


class ModelImpl:
    def __call__(self, __input_batch: InputBatchType) -> List[np.ndarray]:
        target_batch = np.zeros((N_DATAPOINTS, N_CLASSES))
        for i, target_instance in enumerate(target_batch):
            target_instance[i % N_CLASSES] = 1

        return [i for i in target_batch]


class MetricImpl:
    def __init__(self):
        return None

    def reset(self) -> None:
        return None

    def update(
        self, __pred_batch: TargetBatchType, __target_batch: TargetBatchType
    ) -> None:
        return None

    def compute(self) -> Dict[str, Any]:
        return {"metric1": "val1", "metric2": "val2"}



# Copyright 2024, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT

# import component generics from generic.py and specialize them for object detection
# domain

from __future__ import annotations

import os
from typing import (
    Any,
    Callable,
    Iterable,
    List,
    Literal,
    Protocol,
    Union,
    overload,
    runtime_checkable,
)

from typing_extensions import Self, TypeAlias

from maite._internals.protocols import (
    generic as gen,
    image_classification as ic,
    object_detection as od,
    task_aliases as al,
)


@runtime_checkable
class ModelProvider(Protocol):
    def help(self, name: str) -> str:
        """
        Get information about the model such as:

            - Instructions for its use
            - Intended purpose
            - Any academic references

        Parameters
        ----------
        name: str
            The key that can be used to retrieve the model, returned by
            ``~ModelProvider.list_models``, same value used to retrieve the
            Model from ``~ModelProvider.load_model``

        Returns
        -------
        output: str
            The informational text.
        """
        ...

    def list_models(
        self,
        *,
        filter_str: str | List[str] | None = None,
        model_name: str | None = None,
        task: al.TaskName | None = None,
    ) -> Iterable[Any]:
        """
        List models for this provider.

        Parameters
        ----------
        filter_str : str | List[str] | None (default: None)
            A string or list of strings that contain complete or partial names for models.
        model_name : str | None (default: None)
            A string that contain complete or partial names for models.
        task : TaskName | None (default: None)
            A string or list of strings of tasks models were designed for, such as: "image-classification", "object-detection".
        **kwargs : Any
            Any keyword supported by this provider interface.

        Returns
        -------
        Iterable[Any]
            An iterable of model names.

        """
        ...

    @overload
    def load_model(
        self, model_name: str, task: Literal["image-classification"]
    ) -> ic.Model:
        ...

    @overload
    def load_model(
        self, model_name: str, task: Literal["object-detection"]
    ) -> od.Model:
        ...

    def load_model(self, model_name: str, task: al.TaskName) -> al.SomeModel:
        """
        Return a supported model.

        Parameters
        ----------
        model_name : str
            The `model_name` for the model (e.g., "microsoft/resnet-18").
        task : str | None
            The task for the model (e.g., "image-classification"). If None the task will be inferred from the model's interface
        **kwargs : Any
            Any keyword supported by provider interface.

        Returns
        -------
        Model
            A Model object that supports the given task.
        """
        ...


@runtime_checkable
class DatasetProvider(Protocol):
    def help(self, name: str) -> str:
        """
        Get information about the dataset such as instructions for its use,
        intended purpose, and relevant references

        Parameters
        ----------
        name: str
            The key that can be used to retrieve the dataset, returned by
            ``~DatasetProvider.list_datasets``, same value used to retrieve the
            Model from ``~DatasetProvider.load_dataset``

        Returns
        -------
        output: str
            The informational text.
        """
        ...

    def list_datasets(self) -> Iterable[str]:
        """
        List datasets for this provider.

        Parameters
        ----------
        **kwargs : Any
            Any keyword supported by this provider.

        Returns
        -------
        Iterable[Any]
            An iterable of dataset names.

        """
        ...

    @overload
    def load_dataset(
        self,
        *,
        dataset_name: str,
        task: Literal["image-classification"],
        split: str | None = None,
    ) -> ic.Dataset:
        ...

    @overload
    def load_dataset(
        self,
        *,
        dataset_name: str,
        task: Literal["object-detection"],
        split: str | None = None,
    ) -> od.Dataset:
        ...

    def load_dataset(
        self,
        *,
        dataset_name: str,
        task: al.TaskName,
        split: str | None = None,
    ) -> al.SomeDataset:
        """
        Load dataset for a given provider.

        Parameters
        ----------
        dataset_name : str
            Name of dataset.
        task : TaskName | None (default: None)
            A string or list of strings of tasks dataset were designed for, such as: "image-classification", "object-detection".
        split : str | None (default: None)
            A string of split to load, such as: "train", "test", "validation".
            If None, the default split will be returned
        **kwargs : Any
            Any keyword supported by this provider.

        Returns
        -------
        Dataset
            A dataset object that supports the given task.

        """
        ...


@runtime_checkable
class MetricProvider(Protocol):
    def help(self, name: str) -> str:
        """
        Get information about the Metric such as:

            - Instructions for its use
            - Intended purpose
            - Any academic references

        Parameters
        ----------
        name: str
            The key that can be used to retrieve the model, returned by
            ``~MetricProvider.list_metric``, same value used to retrieve the
            Model from ``~MetricProvider.load_metric``

        Returns
        -------
        output: str
            The informational text.
        """
        ...

    def list_metrics(self) -> Iterable[Any]:
        """
        List metrics for this provider.

        Parameters
        ----------
        **kwargs : Any
            Any keyword supported by this provider.

        Returns
        -------
        Iterable[Any]
            An iterable of metric names.

        """
        ...

    @overload
    def load_metric(self, metric_name: str) -> ic.Metric:
        ...

    @overload
    def load_metric(self, metric_name: str) -> od.Metric:
        ...

    def load_metric(self, metric_name: str) -> al.SomeMetric:
        # load_metric must return a supertype of all task-specific metrics
        # since gen.Metric is contravariant wrt type argument, gen.Metric[Union[od.TargetBatchType, ic.TargetBatchType,...]]
        # is gen.Metric[<SuperType of (od.TargetBatchType, ic.TargetBatchType,...)>]
        # which is a *subtype* of all task specific metrics (note the reversal of supertype/subtype relationship)
        # when types are passed through contravariant argument of a generic
        """
        Return a Metric object.

        Parameters
        ----------
        metric_name : str
            The `metric_name` for the metric (e.g., "accuracy").
        **kwargs : Any
            Any keyword supported by this provider.

        Returns
        -------
        Metric
            A Metric object.
        """
        ...


# Why an alias union, rather than a base protocol?  There are some applications where
# the concept of AnyProvider is a meaningful type hint However there is no useful
# declaration that could be assigned to a base type say with only the common `help`
# method as a requirement
AnyProvider: TypeAlias = Union[ModelProvider, MetricProvider, DatasetProvider]


#
# ArtifactHub
#
@runtime_checkable
class ArtifactHubEndpoint(Protocol):
    def __init__(self, path: Any):
        """Endpoints are initialized with a path argument specifying any information necessary to find the source via the endpoint's target api"""
        ...

    def get_cache_or_reload(self) -> str | os.PathLike[str]:
        """Create a local copy of the resource in the cache (if needed) and return a path suitable to locate the `hubconf.py` file"""
        ...

    def update_options(self) -> Self:
        """Update update any cached state used by the endpoint

        API tokens, validation options, etc.
        If none of these apply, the method may simply be a no-op
        """
        ...


ModelEntrypoint: TypeAlias = Callable[..., gen.Model]
DatasetEntrypoint: TypeAlias = Callable[..., gen.Dataset]
MetricEntrypoint: TypeAlias = Callable[..., gen.Metric]
AnyEntrypoint: TypeAlias = Union[ModelEntrypoint, DatasetEntrypoint, MetricEntrypoint]



# Copyright 2024, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT

"""Define aliases over all tasks"""

# We need a module beyond protocols/__init__.py for 'union aliases' because aliases depend on all
# task-specific modules (e.g. 'image_classification.py') and task-specific modules depend
# on __init__.py (through ArrayLike). Thus, introducing these aliases in __init__.py
# results in a circular import). These task-agnostic aliases are meant to be imported
# only in code where such task-specific aliases are not defined.

# We don't define aliases over all *components* here (e.g. 'AnyProvider: TypeAlias = ...', 'AnyEntrypoint: ...')
# because we don't plan on defining component-specific provider modules defined separately from
# consumers of these 'AnyProvider'. We expect the list of component types should be more static
# than the list of tasks

from typing import Any, Literal, Union

from typing_extensions import TypeAlias

from maite._internals.protocols import (
    generic as gen,
    image_classification as ic,
    object_detection as od,
)

# define "Some<ComponentType>" and "Some<PrimitiveType>" TypeAliases
SomeModel: TypeAlias = Union[ic.Model, od.Model]
SomeMetric: TypeAlias = Union[ic.Metric, od.Metric]
SomeDataset: TypeAlias = Union[ic.Dataset, od.Dataset]
SomeDataLoader: TypeAlias = Union[ic.DataLoader, od.DataLoader]
SomeAugmentation: TypeAlias = Union[ic.Augmentation, od.Augmentation]

SomeInputBatchType: TypeAlias = Union[ic.InputBatchType, od.InputBatchType]
SomeTargetBatchType: TypeAlias = Union[ic.TargetBatchType, od.TargetBatchType]
SomeMetadataBatchType: TypeAlias = Union[
    ic.DatumMetadataBatchType, od.DatumMetadataBatchType
]

SomeInputType: TypeAlias = Union[ic.InputType, od.InputType]
SomeTargetType: TypeAlias = Union[ic.TargetType, od.TargetType]
SomeMetadataType: TypeAlias = Union[ic.DatumMetadataType, od.DatumMetadataType]

TaskName: TypeAlias = Literal["object-detection", "image-classification"]

# Note: We could write a single generic with bound parameters, but it wouldn't
# capture coupling between different parameters (e.g. InputType and TargetType
# need to come from same task)

# Write aliases for openly-typed components that are not meant to be directly
# exposed in any overloads signatures, but are nonetheless needed in
# the signature of task-agnostic workflow implementations. We consider this
# aliasing better than directly "Any-typing" arguments to implementations

OpenModelType = gen.Model[Any, Any]
OpenDatasetType = gen.Dataset[Any, Any, Any]
OpenMetricType = gen.Metric[Any]
OpenDataLoaderType = gen.DataLoader[Any, Any, Any]
OpenAugmentationType = gen.Augmentation[Any, Any, Any, Any, Any, Any]
OpenCollateFnType = gen.CollateFn[Any, Any, Any]



# Copyright 2024, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT

from __future__ import annotations

import warnings
from enum import Enum
from typing import (
    Any,
    Callable,
    ClassVar,
    Dict,
    List,
    Literal,
    NamedTuple,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
    get_args as get_type_args,
    overload,
)

from maite._internals.protocols.providers import (
    DatasetProvider,
    MetricProvider,
    ModelProvider,
)
from maite.errors import InternalError, InvalidArgument

ArtifactName = Literal["model", "dataset", "metric"]
AnyProvider = Union[ModelProvider, DatasetProvider, MetricProvider]

_PROVIDER_PROTOCOLS: Tuple[Type[AnyProvider], ...] = tuple(get_type_args(AnyProvider))

# And from that build the enum
# The enum gives us a compact way to map a desired artifact on the proper protocol interface
# name -> tag: ProviderProtocolTag[name]
# tag -> Protocol: PROVIDER_PROTOCOLS[tag.value]
_ProviderProtocolTag = Enum(
    "_ProviderProtocolTag", ((x, i) for i, x in enumerate(get_type_args(ArtifactName)))
)


#
# Provider Registry Internals
#


class _RegisteredProviderInfo(NamedTuple):
    """The value used to store the registered implementation by the registry

    Attributes
    ----------
    impl: Type[AnyProvider]
      The type which has been registered as a provider
    implements:
      A tuple of ProviderProtocolTag's indicating which protocols ``impl`` supports

    Notes
    -----
    Why cache the implements parameter?
        The runtime checks on protocols can be expensive, so we do them once at
        registration time and cache the results to check against at the time when an
        implementation is looked up for use.
    """

    impl: Type[AnyProvider]
    implements: Tuple[_ProviderProtocolTag, ...]


class _ProviderRegistry:
    """The registry for provider implementations
    This class holds the mapping of key names -> provider implementations for use by
    the factories

    Attributes
    ----------
    registered_provider : Dict[str, RegisteredProviderInfo]
        Maps registered names -> provider implementation types

    Methods
    -------
    register_impl(impl_type: Type[AnyProvider], key: str) -> None
        Register a provider under the given key
    get_impl(key: str, enforce_protocol: ArtifactName | None = None) -> Type[AnyProvider]
        Lookup the Provider type associated with a key
    list_registered(enforce_protocol: ArtifactName | None = None) -> List[str]
        List provider keys registered (optionally limit to specific protocol)

    """

    registered_providers: ClassVar[Dict[str, _RegisteredProviderInfo]] = {}

    def __init__(self, *args, **kwargs):
        # Error on __init__ no need to instantiate the class Technically, this would be
        # "safe" in that the instance would hold a reference to the dict defined above,
        # but there is no need to allow/encourage doing that
        raise InternalError(
            "The ProviderRegistry functionality should be accessed "
            "through the class itself, do not instantiate this type"
        )

    @classmethod
    def register_impl(cls, impl_type: Type[AnyProvider], key: str):
        """Register a Provider implementation

        Parameters
        ----------
        impl_type : Type[AnyProvider]
           Some type implementing at least one of the provider protocols
        key : str
           The string name that can be used to find that provider implementation

        Examples
        --------
        >>> _ProviderRegistry.register_impl(MyProviderType, key="my_provider")

        """
        if key in cls.registered_providers:
            # Warn on key overwrite, we allow this because there are legitimate uses for
            # the pattern, extending a native provider to add some special checks
            # specific to that user's org/standards/security protocols for examples.  We
            # still warn so that someone knows they have done this. They can filter the
            # warning around the registration site if they wish to have a warning-free
            # experience
            warnings.warn(
                f"Attempting to register provider {impl_type.__name__} under "
                f"name {key} which will overwrite the existing provider "
                f"{cls.registered_providers[key].impl.__name__} registered under "
                "that name"
            )
        implements_protocols = tuple(
            _ProviderProtocolTag(i)
            for i, p in enumerate(_PROVIDER_PROTOCOLS)
            if isinstance(impl_type, p)
        )
        if not implements_protocols:
            # Error when the implementation does not satisfy any protocols. They can use
            # a provider-like type without registration if they have a very odd
            # situation that requires it, but passing the provider directly should be
            # possible anywhere a factory-key is also acceptable, checks at those
            # interact should be static.  Explicitly disallow registration of incomplete
            # provider implementations to prevent them using one which is ill-formed
            # unless they know what they are doing
            raise InvalidArgument(
                f"Attempting to register a provider {impl_type.__name__} "
                f"which does not satisfy any of the protocols "
                f"{_PROVIDER_PROTOCOLS}."
            )
        cls.registered_providers[key] = _RegisteredProviderInfo(
            impl_type, implements_protocols
        )

    @classmethod
    def list_registered(cls, enforce_protocol: ArtifactName | None = None) -> List[str]:
        """List the implementations registered which satisfy requirements to be a provider for a particular artifact"""
        if enforce_protocol:
            enforce_tag = _ProviderProtocolTag[enforce_protocol]
            return [
                k
                for k in cls.registered_providers.keys()
                if enforce_tag in cls.registered_providers[k].implements
            ]

        return list(cls.registered_providers.keys())

    @classmethod
    def get_impl(
        cls,
        key: str,
        enforce_protocol: ArtifactName | None = None,
    ) -> Type[AnyProvider]:
        """Get the type associated with the provider key

        Parameters
        ----------
        key : str
           The name used to register the provider
        enforce_protocol : ArtifactName, default = None
            The default behavior, when the value is None, is to not enforce a specific
            artifact. Any registered type will implement at least one complete provider
            protocol.

        Returns
        -------
        Type[AnyProvider]
           A type (class) that implements at least one of the provider protocols. If
           restricted with ``enforce_protocol`` registered types which do not implement
           that specific artifact protocol will be considered invalid.

        Raises
        ------
        InvalidArgument
           If the key is not registered, or the associated type does not implement the
           protocol when using ``enforce_protocol``

        """
        impl_info = cls.registered_providers.get(key, None)
        if impl_info is None:
            registered_msg = "\n\t".join(cls.list_registered(enforce_protocol))
            raise InvalidArgument(
                f"Unable to find provider registered for key:  {key}, "
                f"the following are registered: \n\t{registered_msg}\n"
            )

        if enforce_protocol:
            enforce_tag = _ProviderProtocolTag[enforce_protocol]
            if enforce_tag not in impl_info.implements:
                implements_msg = "\n\t".join(i.name for i in impl_info.implements)
                raise InvalidArgument(
                    f"The key '{key}' does not implement the interface "
                    f"for {enforce_tag.name} providers. The associated "
                    f"type {impl_info.impl.__name__} implements the "
                    f"following providers:\n {implements_msg}"
                )

        return impl_info.impl


#
# Registration Methods
#

PT = TypeVar("PT", bound=AnyProvider)


@overload
def register_provider(
    provider_type: Type[PT], *, key: Optional[str] = None
) -> Type[PT]:
    ...


@overload
def register_provider(
    provider_type: None,
) -> Callable[[Type[PT]], Type[PT]]:
    ...


@overload
def register_provider(*, key: Optional[str] = None) -> Callable[[Type[PT]], Type[PT]]:
    ...


@overload
def register_provider(provider_type: Type[PT]) -> Type[PT]:
    ...


def register_provider(
    provider_type: Optional[Type[PT]] = None, *, key: Optional[str] = None
) -> Callable[[Type[PT]], Type[PT]] | Type[PT]:
    """
    Register a provider.

    Parameters
    ----------
    provider_type : Type[AnyProvider]
       A class which implements one of the provider protocols. The registry will enforce this requirement dynamically.
    key : str, default = None
       The key this implementation will be associated with in the registry.

    Returns
    -------
    AnyProvider
       The registered implementation, unchanged.

    Raises
    ------
    InvalidAugmentError
       If the ``provider_type`` does not implement at least one of the provider protocols
       If a key is not provided, and the class does not define the PROVIDER_REGISTRY_KEY
       attribute.

    Examples
    --------
    This utility can be used in several ways, lets begin with a simple, valid Provider type

    >>> class SimpleProvider:
    ...     def __init__(self, data: Dataset, name: str):
    ...         self.dataset = data
    ...         self.name = name
    ...     def help(self, name):
    ...         return "This provider only has one dataset, there is no information about it"
    ...     def list_datasets(self):
    ...         return [self.name]
    ...     def load_dataset(self, *, name: str, task: TaskName, split: str):
    ...         return self.dataset

    The type can be registered with an explicit key, by calling the function

    >>> register_provider(SimpleProvider, key="simple") # type: ignore

    Classes which define the ``PROVIDER_REGISTRY_KEY`` attribute can be registered
    without an explicit ``key`` keyword argument

    >>> class SimpleProviderWithKeyDef(SimpleProvider):
    ...     PROVIDER_REGISTRY_KEY = "another_simple"
    >>> register_provider(SimpleProviderWithKeyDef) # type: ignore

    This function can also act as a decorator on the class definition

    >>> @register_provider # type: ignore
    ... class SimpleProviderWithKeyDefDecorated(SimpleProvider):
    ...     PROVIDER_REGISTRY_KEY = "decorated_simple"

    Or without implying the key from the class attribute

    >>> @register_provider(key="key_decorated") # type: ignore
    >>> class SimpleProviderDecorated(SimpleProvider):
    ...     ...

    All of these implement the DatasetProvider protocol so they will be registered with
    the factory and associated with "dataset" protocol.

    >>> list_providers(enforce_protocol='dataset')
    ['simple', 'another_simple', 'decorated_simple']

    However, they will not be associated with other artifact types as they don't
    implement the necessary interface.

    >>> list_providers(enforce_protocol='model')
    []
    """

    def _registration_helper(
        provider_type: Type[PT], *, key: Optional[str] = None
    ) -> Type[PT]:
        if key is None:
            key = getattr(provider_type, "PROVIDER_REGISTRY_KEY", None)

        if key is None:
            raise InvalidArgument(
                "Registering a provider implementation without a `key` keyword argument requires the type define the class attribute PROVIDER_REGISTRY_KEY"
            )

        _ProviderRegistry.register_impl(provider_type, key)
        return provider_type

    def wrap(provider_t):
        return _registration_helper(provider_t, key=key)

    if provider_type is None:
        # called like @register_provider()
        return wrap

    # called like @register_provider
    return wrap(provider_type)


#
# Query the registry
#


def list_providers(enforce_protocol: ArtifactName | None = None) -> List[str]:
    """
    List registered provider keys.

    Parameters
    ----------
    enforce_protocol : ArtifactName, default None
       When provided results will be filtered to registered providers which can provide
       the specified artifact type. The default behavior is to list all providers.

    Returns
    -------
    List[str]
       A list of keys which be used to look up a provider which is able to provide the
       specified artifact type.

    Examples
    --------
    >>> list_providers()
    ... [...] # contains native providers only
    >>> class MyProvider:
    ...    ...
    >>> register_provider(MyProvider, key='my_provider') # type: ignore
    >>> 'my_provider' in list_providers()
    True
    """
    return _ProviderRegistry.list_registered(enforce_protocol)


def get_provider_type(
    key: str, enforce_protocol: ArtifactName | None = None
) -> Type[AnyProvider]:
    """
    Get the type from the registry based on its key.

    Parameters
    ----------
    key : str
       The key associated with the desired provider type in the registry.
    enforce_protocol : ArtifactName, default None
       When provided, an error is raised if the provider does is not able to provide the specified artifact type.

    Returns
    -------
    Type[AnyProvider]
       The provider type registered with the given key.

    Raises
    ------
    InvalidArgumentError
       When there is no suitable provider registered with the given key.

    Examples
    --------
    >>> from typing import List
    >>> class MyProvider:
    ...    ...
    >>> register_provider(MyProvider, key='my_provider') #type: ignore
    >>> get_provider_type('my_provider') is MyProvider
    True
    """
    return _ProviderRegistry.get_impl(key, enforce_protocol)


def create_provider(
    key: str,
    enforce_protocol: ArtifactName | None = None,
    provider_args: Tuple[Any, ...] | None = None,
    provider_kwargs: Dict[str, Any] | None = None,
) -> AnyProvider:
    """
    Create and return an instance of the specified provider type.

    Parameters
    ----------
    key : str
       The key associated with the desired provider type in the registry.
    enforce_protocol : ArtifactName, default None
       When provided, an error is raised if the provider does is not able to provide the specified artifact type.
    provider_args : Tuple[Any, ...], default None
       Args to forward to the provider type's __init__.
    provider_kwargs : Dict[str, Any], default None
       Kwargs to forward to the provider type's __init__.

    Returns
    -------
    AnyProvider
       An instance of the provider registered with the given key.

    Raises
    ------
    InvalidArgument
        When there is no suitable provider registered with the given key.

    Notes
    -----
    This method will invoke the type's __init__ which may raise other exceptions than
    those listed here.

    Examples
    --------
    >>> @register_provider(key='my_provider') # type: ignore
    ... class MyProvider:
    ...    def __init__(self, arg1, arg2):
    ...        self.arg1 = arg1
    ...        self.arg2 = arg2
    >>> p = create_provider(key='my_provider', provider_args=('a', 'b'))
    >>> p.arg1 # type: ignore
    'a'
    >>> p.arg2 # type: ignore
    'b'
    """
    _args = provider_args or ()
    _kwargs = provider_kwargs or {}
    return get_provider_type(key, enforce_protocol)(*_args, **_kwargs)



# Copyright 2024, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT

import pytest

from maite.testing.docs import validate_docstring
from maite.testing.project import get_public_symbols, import_public_symbols
from maite.testing.pyright import list_error_messages, pyright_analyze
from tests import module_scan

# Generates a string that imports all symbols from the maite's public API.
# TODO: Make this a convenience function exposed via our public API?
preamble = "\n".join(
    [
        "from {} import {}".format(*x["name"].rsplit(".", maxsplit=1))
        for x in get_public_symbols(module_scan("maite"))
        if x["category"] in {"module", "function", "class", "type alias"}
    ]
)

all_funcs_and_classes = list(
    import_public_symbols(
        module_scan("maite"),
        skip_module_not_found="pytest-skip",
    ),
)

PYRIGHT_SCAN_RESULTS = []
FUNCS_TO_SCAN = [
    obj
    for obj in all_funcs_and_classes
    if obj.__doc__ is not None and obj is not pyright_analyze
]
for obj, scan in zip(
    FUNCS_TO_SCAN,
    pyright_analyze(
        *FUNCS_TO_SCAN,
        scan_docstring=True,
        report_unnecessary_type_ignore_comment=True,
        preamble=preamble,
    ),
):
    PYRIGHT_SCAN_RESULTS.append([obj, scan])


@pytest.mark.parametrize("obj, scan", PYRIGHT_SCAN_RESULTS)
def test_docstrings_scan_clean_via_pyright(obj, scan):
    if scan["summary"]["errorCount"] != 0:
        raise ValueError("\n" + "\n".join(list_error_messages(scan)))


@pytest.mark.parametrize("obj", all_funcs_and_classes)
def test_docstrings_adhere_to_numpydoc(obj):
    results = validate_docstring(obj, ignore=("SA01", "ES01"))

    if results["error_count"] > 0:
        errs = []
        for err_key, err_val in results["errors"].items():
            errs.append(f"{err_key}: {err_val}")
        raise AssertionError(
            "\n"
            + f"{obj.__module__}.{obj.__name__} docstring doesn't adhere to numpydoc"
            + "\n"
            + "\n".join(errs)
        )



# Copyright 2024, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT

from typing import Any, Protocol, runtime_checkable


# define ArrayLike protocol
@runtime_checkable
class ArrayLike(Protocol):
    """
    A protocol for an array-like object.

    Examples
    --------

    Arrays like NumPy NDArray objects are `ArrayLike` along
    with PyTorch and JAX tensors.

    >>> import numpy as np
    >>> from maite.protocols import ArrayLike
    >>> array_like: ArrayLike = np.ones((3, 224, 224))
    >>> isinstance(array_like, ArrayLike)
    True
    """

    def __array__(self) -> Any:
        ...


# define minimal DatumMetadata protocol class
@runtime_checkable
class DatumMetadata(Protocol):
    @property
    def uuid(self) -> int:
        ...



# Copyright 2024, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT

from dataclasses import dataclass
from typing import Final, Literal, Protocol

from typing_extensions import TypeAlias, TypedDict

__all__ = [
    "CONSTANT",
    "Alias",
    "AClass",
    "ADataClass",
    "ATypedDict",
    "AProtocol",
    "a_func",
]

CONSTANT: Final = 22

Alias: TypeAlias = Literal["a", "b"]


def a_func(x: int) -> None:
    ...


@dataclass
class ADataClass:
    x: int


class AClass:
    @staticmethod
    def static_meth() -> None:
        ...

    def meth(self) -> None:
        ...


class ATypedDict(TypedDict):
    x: int


class AProtocol(Protocol):
    def meth(self) -> None:
        ...



# Copyright 2024, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT

from maite._internals.protocols import generic as gen
from maite.workflows import evaluate, predict


def test_simple_component_structural(
    od_simple_augmentation,
    od_simple_dataset,
    od_simple_dataloader,
    od_simple_model,
    od_simple_metric,
):
    # verify types pass isinstance checks

    assert isinstance(od_simple_augmentation, gen.Augmentation)
    assert isinstance(od_simple_metric, gen.Metric)
    assert isinstance(od_simple_dataset, gen.Dataset)
    assert isinstance(od_simple_dataloader, gen.DataLoader)
    assert isinstance(od_simple_model, gen.Model)


def test_simple_od_evaluate(
    od_simple_augmentation,
    od_simple_dataset,
    od_simple_dataloader,
    od_simple_model,
    od_simple_metric,
):
    # Run types through "evaluate" workflow

    evaluate(
        model=od_simple_model,
        dataloader=od_simple_dataloader,
        metric=od_simple_metric,
        augmentation=od_simple_augmentation,
    )

    evaluate(
        model=od_simple_model,
        dataset=od_simple_dataset,
        metric=od_simple_metric,
        augmentation=od_simple_augmentation,
    )


def test_simple_od_predict(
    od_simple_augmentation,
    od_simple_dataset,
    od_simple_dataloader,
    od_simple_model,
    od_simple_metric,
):
    # Run types through "predict" workflow

    predict(
        model=od_simple_model,
        dataloader=od_simple_dataloader,
        augmentation=od_simple_augmentation,
    )

    predict(
        model=od_simple_model,
        dataset=od_simple_dataset,
        augmentation=od_simple_augmentation,
    )



# Copyright 2024, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT

"""
Utilities for validating argument types and values that raise clear, standardized error
messages.
"""

from maite._internals.validation import (
    chain_validators,
    check_domain,
    check_one_of,
    check_type,
)

__all__ = ["check_type", "check_domain", "check_one_of", "chain_validators"]



# Copyright 2024, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT

import inspect
import json
import os
import re
import shutil
import subprocess
import tempfile
import textwrap
from collections import Counter, defaultdict
from contextlib import contextmanager
from pathlib import Path
from typing import Any, DefaultDict, Dict, List, Optional, Union

from typing_extensions import Literal, NotRequired, TypedDict


def notebook_to_py_text(path_to_nb: Path) -> str:
    try:
        import jupytext
    except ImportError:
        raise ImportError(
            "`jupytext` must be installed in order to run pyright on jupyter notebooks."
        )
    ntbk = jupytext.read(path_to_nb, fmt="ipynb")
    return jupytext.writes(ntbk, fmt=".py")


class Summary(TypedDict):
    filesAnalyzed: int
    errorCount: int
    warningCount: int
    informationCount: int
    timeInSec: float


class LineInfo(TypedDict):
    line: int
    character: int


class Range(TypedDict):
    start: LineInfo
    end: LineInfo


class Diagnostic(TypedDict):
    file: str
    severity: Literal["error", "warning", "information"]
    message: str
    range: Range
    rule: NotRequired[str]


class PyrightOutput(TypedDict):
    """The schema for the JSON output of a pyright scan"""

    # # doc-ignore: NOQA
    version: str
    time: str
    generalDiagnostics: List[Diagnostic]
    summary: Summary


_found_path = shutil.which("pyright")
PYRIGHT_PATH = Path(_found_path) if _found_path else None
del _found_path


@contextmanager
def chdir():
    with tempfile.TemporaryDirectory() as tmpdirname:
        old_dir = os.getcwd()  # get current working directory (cwd)
        try:
            os.chdir(tmpdirname)  # change cwd to the temp-directory
            yield Path(tmpdirname)  # yields control to the test to be run
        finally:
            os.chdir(old_dir)


# `docstring_re` is derived from https://github.com/python/cpython/blob/main/Lib/doctest.py
# which is free for copy/reuse under GPL license
#
# This regular expression is used to find doctest examples in a
# string.  It defines two groups: `source` is the source code
# (including leading indentation and prompts); `indent` is the
# indentation of the first (PS1) line of the source code; and
# `want` is the expected output (including leading indentation).
docstring_re = re.compile(
    r"""
    # Source consists of a PS1 line followed by zero or more PS2 lines.
    (?P<source>
        (?:^(?P<indent> [ ]*) >>>    .*)    # PS1 line
        (?:\n           [ ]*  \.\.\. .*)*)  # PS2 lines
    \n?
    """,
    re.MULTILINE | re.VERBOSE,
)


def get_docstring_examples(doc: str) -> str:
    prefix = ">>> "

    # contains input lines of docstring examples with all indentation
    # and REPL markers removed
    src_lines: List[str] = []

    for source, indent in docstring_re.findall(doc):
        source: str
        indent: str
        for line in source.splitlines():
            src_lines.append(line[len(indent) + len(prefix) :])
        src_lines.append("")  # newline between blocks
    return "\n".join(src_lines)


def rst_to_code(src: str) -> str:
    """
    Consumes rst-formatted text like::

       lopsem est decorum

       .. code-block:: python
          :caption: blah
          :name: bark bark

          import math
          x = 1+1

       foorbarius isbarfooium

       .. code-block:: pycon
          :caption: blah

          >>> print("hi")
          hi
          >>> 2+1

    and returns the string::

       '''
       import math
       x = 1+1


       print("hi")

       2+1
       '''

    """
    block: Optional[List[str]] = None  # lines in code block
    indentation: Optional[str] = None  # leading whitespace before .. code-block
    preamble: Optional[str] = None  # python or pycon
    n = -float("inf")  # line no in code block
    blocks: List[str] = []  # respective code blocks, each ready for processing

    def add_block(
        block: Optional[List[str]],
        preamble: Optional[str],
        blocks: List[str],
    ):
        if block:
            block_str = "\n".join(block) + "\n"
            assert preamble
            if "pycon" in preamble:
                blocks.append(get_docstring_examples(block_str))
            else:
                blocks.append(textwrap.dedent(block_str))

    for line in src.splitlines():
        n += 1
        # 0 <= n: if within code block

        if line.strip().startswith(".. code-block:: py"):
            # Entering python/pycon code block
            add_block(block, preamble, blocks)
            n = -1
            block = []
            indentation = line.split("..")[0] + " " * 3
            preamble = line.split("::")[-1].strip()
            continue

        if n < 0:
            # outside of code block
            continue

        assert indentation is not None
        assert block is not None

        if not (line.startswith(indentation) or not line.strip()):
            # encountering non-empty line that isn't within
            # minimum indentation leaves the code block
            add_block(block, preamble, blocks)
            block = None
            n = -float("inf")
            continue

        if n == 0:
            # first line of code block is either empty or a directive
            stripped = line.strip()
            if not stripped:
                continue

            if line.startswith(indentation):
                if stripped.startswith(":"):
                    n = -1
                    # skip directive, act as if we are at top of code block
                    continue

        block.append(line)

    add_block(block, preamble, blocks)
    return "\n".join(blocks)


def md_to_code(src: str) -> str:
    """
    Consumes markdown-formatted text like::

       lopsem est decorum

       ```python
       import math
       x = 1+1
       ```

       foorbarius isbarfooium

       ```pycon
        >>> print("hi")
        hi
        >>> 2+1
        ```

    and returns the string::

       '''
       import math
       x = 1+1


       print("hi")

       2+1
       '''

    """
    block: Optional[List[str]] = None  # lines in code block
    preamble: Optional[str] = None  # python or pycon
    blocks: List[str] = []  # respective code blocks, each ready for processing

    # inside
    # ````
    # <literal markdown>
    # ````
    in_literal_block: bool = False

    # inside
    # ```py[con][thon]
    # <code>
    # ```
    in_code_block: bool = False

    def add_block(
        block: Optional[List[str]],
        preamble: Optional[str],
        blocks: List[str],
    ):
        if block:
            block_str = "\n".join(block) + "\n"
            assert preamble
            if "pycon" == preamble:
                blocks.append(get_docstring_examples(block_str))
            elif "python" == preamble:
                blocks.append(textwrap.dedent(block_str))
            else:
                # unknown block
                pass

    for line in src.splitlines():
        stripped = line.strip()

        if stripped == "`" * 4:
            in_literal_block = not in_literal_block

        if not in_literal_block and stripped.startswith("```py"):
            # Entering python/pycon code block
            add_block(block, preamble, blocks)
            assert not in_code_block, line
            in_code_block = True
            block = []
            preamble = line.split("`" * 3)[-1].strip()
            continue

        if not in_code_block:
            # outside of code block
            continue

        assert block is not None

        if not in_literal_block and stripped == "`" * 3:
            # encountering ``` leaves the code block
            add_block(block, preamble, blocks)
            block = None
            in_code_block = False
            continue

        block.append(line)

    add_block(block, preamble, blocks)
    return "\n".join(blocks)


def pyright_analyze(
    *code_objs_and_or_paths: Any,
    pyright_config: Optional[Dict[str, Any]] = None,
    scan_docstring: bool = False,
    path_to_pyright: Union[Path, None] = PYRIGHT_PATH,
    preamble: str = "",
    python_version: Optional[str] = None,
    report_unnecessary_type_ignore_comment: Optional[bool] = None,
    type_checking_mode: Optional[Literal["basic", "strict"]] = None,
) -> List[PyrightOutput]:
    r"""
    Scan a Python object, docstring, or file with pyright.

    The following file formats are supported: `.py`, `.rst`, `.md`, and `.ipynb`.

    Some common pyright configuration options are exposed via this function for
    convenience; a full pyright JSON config can be specified to completely control
    the behavior of pyright.

    This function requires that pyright is installed and can be run from the command
    line [1]_.

    Parameters
    ----------
    *code_objs_and_or_paths : Any
        A function, module-object, class, or method to scan. Or, a path to a file
        to scan. Supported file formats are `.py`, `.rst`, `.md, and `.ipynb`.

        Specifying a directory is permitted, but only `.py` files in that directory
        will be scanned. All files will be copied to a temporary directory before being
        scanned.

    pyright_config : None | dict[str, Any]
        A JSON configuration for pyright's settings [2]_.

    scan_docstring : bool, optional (default=False), keyword-only
        If `True` pyright will scan the docstring examples of the specified code object,
        rather than the code object itself.

        Example code blocks are expected to have the doctest format [3]_.

    path_to_pyright : Path, optional, keyword-only
        Path to the pyright executable (see installation instructions: [4]_).
        Defaults to `shutil.where('pyright')` if the executable can be found.

    preamble : str, optional (default=''), keyword-only
        A "header" added to the source code that will be scanned. E.g., this can be
        useful for adding import statements.

    python_version : Optional[str], keyword-only
        The version of Python used for this execution environment as a string in the
        format "M.m". E.g., "3.9" or "3.7".

    report_unnecessary_type_ignore_comment : Optional[bool], keyword-only
        If `True` specifying `# type: ignore` for an expression that would otherwise
        not result in an error will cause pyright to report an error.

    type_checking_mode : Optional[Literal["basic", "strict"]], keyword-only
        Modifies pyright's default settings for what it marks as a warning verses an
        error. Defaults to 'basic'.

    Returns
    -------
    List[Dict[str, Any]]  (In one-to-one correspondence with `code_objs_and_or_paths`)
        The JSON-decoded results of the scan [3]_.
            - version: str
            - time: str
            - generalDiagnostics: List[DiagnosticDict] (one entry per error/warning)
            - summary: SummaryDict

        See Notes for more details.

    Notes
    -----
    When supplying a single .rst file, code blocks demarcated by
    `.. code-block:: py[thon,con]` are parsed and used to populate a single temporary
    .py file that pyright will scan.

    `SummaryDict` consists of:
        - filesAnalyzed: int
        - errorCount: int
        - warningCount: int
        - informationCount: int
        - timeInSec: float

    `DiagnosticDict` consists of:
        - file: str
        - severity: Literal["error", "warning", "information"]
        - message: str
        - range: _Range
        - rule: NotRequired[str]

    References
    ----------
    .. [1] https://github.com/microsoft/pyright/blob/aad650ec373a9894c6f13490c2950398095829c6/README.md#command-line
    .. [2] https://github.com/microsoft/pyright/blob/main/docs/configuration.md
    .. [3] https://docs.python.org/3/library/doctest.html
    .. [4] https://github.com/microsoft/pyright/blob/main/docs/command-line.md#json-output

    Examples
    --------
    Here pyright will record an error when scan a function that attempts to add a
    string-annotated variable to an integer.

    >>> from maite.testing.pyright import pyright_analyze
    >>> def f(x: str):
    ...     return 1 + x
    >>> pyright_analyze(f)[0]
    {'version': ..., 'time': ..., 'generalDiagnostics': [{'file': ..., 'severity': ..., 'message': 'Operator "+" not supported for types "Literal[1]" and "str"', 'range': {'start': {'line': ..., 'character': ...}, 'end': {'line': ..., 'character': ...}}, 'rule': ...}], 'summary': {'filesAnalyzed': ..., 'errorCount': 1, 'warningCount': 0, 'informationCount': 0, 'timeInSec': ...}}

    Whereas this function scans "clean".

    >>> def g(x: int) -> int:
    ...     return 1 + x
    >>> pyright_analyze(g)[0]
    {'version': ..., 'time': ..., 'generalDiagnostics': ..., 'summary': {'filesAnalyzed': ..., 'errorCount': 0, 'warningCount': 0, 'informationCount': 0, 'timeInSec': ...}}

    All imports must occur within the context of the scanned-object, or the imports can
    be specified in a preamble. For example, consider the following

    >>> import math  # import statement is not be in scope of `f`
    >>> def f():
    ...     math.acos(1)
    >>> pyright_analyze(f)[0]["summary"]["errorCount"]
    1

    We can add a 'preamble' do that the `math` module is imported.

    >>> pyright_analyze(f, preamble="import math")[0]["summary"]["errorCount"]
    0

    Scanning a function's docstring.

    >>> def plus_1(x: int):
    ...     '''
    ...     Examples
    ...     --------
    ...     >>> from mylib import plus_1
    ...     >>> plus_1('2')  # <- pyright_analyze will catch typo (str instead of int)
    ...     3
    ...     '''
    ...     return x + 1
    >>> pyright_analyze(plus_1, scan_docstring=True)[0]["summary"]["errorCount"]
    1

    Fixing the docstring issue

    >>> def plus_1(x: int):
    ...     '''
    ...     Examples
    ...     --------
    ...     >>> from mylib import plus_1
    ...     >>> plus_1(2)
    ...     3
    ...     '''
    ...     return x + 1
    >>> pyright_analyze(plus_1, scan_docstring=True)[0]["summary"]["errorCount"]
    0
    """
    if path_to_pyright is None:  # pragma: no cover
        raise ModuleNotFoundError(
            "`pyright` was not found. It may need to be installed."
        )
    if not path_to_pyright.is_file():
        raise FileNotFoundError(
            f"`path_to_pyright` – {path_to_pyright} – doesn't exist."
        )
    if not pyright_config:
        pyright_config = {}

    if python_version is not None:
        pyright_config["pythonVersion"] = python_version

    if report_unnecessary_type_ignore_comment is not None:
        pyright_config[
            "reportUnnecessaryTypeIgnoreComment"
        ] = report_unnecessary_type_ignore_comment

    if type_checking_mode is not None:
        pyright_config["typeCheckingMode"] = type_checking_mode

    sources: List[Optional[str]] = []
    code_objs_and_or_paths_resolved = []
    for code_or_path in code_objs_and_or_paths:
        if scan_docstring and (
            isinstance(code_or_path, (Path, str))
            or getattr(code_or_path, "__doc__") is None
        ):
            raise ValueError(
                "`scan_docstring=True` can only be specified when `code_or_path` is an "
                "object with a `__doc__` attribute that returns a string."
            )

        if isinstance(code_or_path, str):
            code_or_path = Path(code_or_path)

        if isinstance(code_or_path, Path):
            code_or_path = code_or_path.resolve()

            if not code_or_path.exists():
                raise FileNotFoundError(
                    f"Specified path {code_or_path} does not exist. Cannot be scanned by pyright."
                )

            if code_or_path.suffix == ".rst":
                source = rst_to_code(code_or_path.read_text("utf-8"))
            elif code_or_path.suffix == ".md":
                source = md_to_code(code_or_path.read_text("utf-8"))
            elif code_or_path.suffix == ".ipynb":
                source = notebook_to_py_text(code_or_path)
            elif code_or_path.is_file() and code_or_path.suffix != ".py":
                raise ValueError(
                    f"{code_or_path}: File type {code_or_path.suffix} not supported by "
                    "`pyright_analyze`."
                )
            else:
                source = None
        else:
            if preamble and not preamble.endswith("\n"):
                preamble = preamble + "\n"
            if not scan_docstring:
                source = preamble + textwrap.dedent((inspect.getsource(code_or_path)))
            else:
                docstring = inspect.getdoc(code_or_path)
                assert docstring is not None
                source = preamble + get_docstring_examples(docstring)
        sources.append(source)
        code_objs_and_or_paths_resolved.append(code_or_path)

    with chdir():
        cwd = Path.cwd()

        for n, (source, code_or_path) in enumerate(
            zip(sources, code_objs_and_or_paths_resolved)
        ):
            target_dir = cwd / str(n)

            if source is not None:
                target_dir.mkdir()
                file_ = target_dir / f"{getattr(code_or_path, '__name__', 'source')}.py"
                file_.write_text(source, encoding="utf-8")
            else:
                file_ = Path(code_or_path).absolute()
                if file_.is_dir():
                    file_ = shutil.copytree(file_, target_dir / "SCAN_DIR")
                elif file_ != cwd:
                    target_dir.mkdir()
                    file_ = shutil.copy(file_, target_dir / file_.name)

        config_path = cwd / "pyrightconfig.json"

        if pyright_config:
            config_path.write_text(json.dumps(pyright_config))

        proc = subprocess.run(
            [str(path_to_pyright.absolute()), str(cwd.absolute()), "--outputjson"],
            cwd=cwd,
            encoding="utf-8",
            text=True,
            capture_output=True,
        )
        try:
            scan: PyrightOutput = json.loads(proc.stdout)
        except Exception as e:  # pragma: no cover
            print(proc.stdout)
            raise e

    out = scan["generalDiagnostics"]
    diagnostics_by_file: DefaultDict[int, List[Diagnostic]] = defaultdict(list)

    for item in out:
        file_str = item["file"]
        if "SCAN_DIR" in file_str:
            name = Path(file_str).name
            target_dir = file_str.find("SCAN_DIR")
            file_path = Path(file_str[:target_dir])
        else:
            file_path_all = Path(file_str)
            name = file_path_all.name
            file_path = file_path_all.parent

        file_index = int(file_path.name)
        diagnostic = item.copy()
        diagnostic["file"] = name
        diagnostics_by_file[file_index].append(diagnostic)

    results: List[PyrightOutput] = []

    for n in range(len(code_objs_and_or_paths)):
        severities = Counter(d["severity"] for d in diagnostics_by_file[n])
        summary = Summary(
            filesAnalyzed=1,
            errorCount=severities["error"],
            warningCount=severities["warning"],
            informationCount=severities["information"],
            timeInSec=scan["summary"]["timeInSec"],
        )
        results.append(
            PyrightOutput(
                version=scan["version"],
                time=scan["time"],
                generalDiagnostics=diagnostics_by_file[n],
                summary=summary,
            )
        )
    return results


def list_error_messages(results: PyrightOutput) -> List[str]:
    """A convenience function that returns a list of error messages reported by pyright.

    Parameters
    ----------
    results : PyrightOutput
        The results of pyright_analyze.

    Returns
    -------
    list[str]
        A list of error messages.
    """
    # doc-ignore: EX01 SA01 GL01
    return [
        f"(line start) {e['range']['start']['line']}: {e['message']}"
        for e in results["generalDiagnostics"]
        if e["severity"] == "error"
    ]



# Copyright 2024, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT

import logging
import os
import tempfile
from typing import Iterable

import pytest


@pytest.fixture()
def cleandir() -> Iterable[str]:
    """
    A pytest fixture that runs a test in a temporary directory as the current working directory.

    This is helpful for running tests that require file I/O that could pollute
    local directories. File cleanup is handled automatically.

    Yields
    ------
    tmpdirname : str
       Temporary directory name that will be removed at fixture teardown

    Examples
    --------
    This assumes that `cleandir` has been imported by the test suite's
    `conftest.py` file.

    >>> import pytest
    >>> @pytest.mark.usefixtures("cleandir")
    ... def test_writes_some_file():
    ...     from pathlib import Path
    ...     Path("dummy.txt").touch()  # file will be written to a tmp dir
    """
    with tempfile.TemporaryDirectory() as tmpdirname:
        old_dir = os.getcwd()  # get current working directory (cwd)
        os.chdir(tmpdirname)  # change cwd to the temp-directory
        yield tmpdirname  # yields control to the test to be run
        os.chdir(old_dir)
        logging.shutdown()



# Copyright 2024, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT

from importlib import import_module

import pytest
from pytest import Config

from tests import all_dummy_subpkgs


def test_version():
    import maite

    assert isinstance(maite.__version__, str)
    assert maite.__version__
    assert "unknown" not in maite.__version__


def test_xfail_strict(pytestconfig: Config):
    # Our test suite's xfail must be configured to strict mode
    # in order to ensure that contrapositive tests will actually
    # raise.
    assert pytestconfig.getini("xfail_strict") is True


@pytest.mark.parametrize("subpkg_name", all_dummy_subpkgs)
def test_dummy_projects_installed(subpkg_name: str):
    import_module(f"maite_dummy.{subpkg_name}")



# Copyright 2024, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT

from pathlib import Path
from typing import List

import pytest
from pytest import param

from maite.errors import InvalidArgument
from maite.testing.project import (
    CompletenessSection,
    ModuleScan,
    ModuleScanResults,
    Symbol,
    SymbolCounts,
    get_public_symbols,
    import_public_symbols,
)
from maite.testing.pyright import Summary
from tests import module_scan

ParameterSet = type(param("s"))


def test_bad_module_name():
    scanner = ModuleScan()

    with pytest.raises(
        ModuleNotFoundError,
        match=r"No files were found to analyze in association with "
        "module `NOT_VALID_MODULE`",
    ):
        scanner("NOT_VALID_MODULE")


def test_bad_path_to_pyright():
    scanner = ModuleScan()
    with pytest.raises(
        FileNotFoundError,
        match=r"`path_to_pyright` – bad_path – doesn't exist",
    ):
        scanner("maite", path_to_pyright=Path("bad_path"))


def test_scan_caching():
    scanner = ModuleScan()
    assert scanner.cache_info().currsize == 0
    assert scanner.cache_info().misses == 0
    assert scanner.cache_info().hits == 0

    scanner("maite")
    assert scanner.cache_info().currsize == 1
    assert scanner.cache_info().misses == 1
    assert scanner.cache_info().hits == 0

    scanner("maite")
    scanner("maite")
    assert scanner.cache_info().currsize == 1
    assert scanner.cache_info().misses == 1
    assert scanner.cache_info().hits == 2

    scanner.clear_cache()
    assert scanner.cache_info().currsize == 0
    assert scanner.cache_info().misses == 0
    assert scanner.cache_info().hits == 0


def test_known_scan():
    results = module_scan("maite")["typeCompleteness"]
    assert results["packageName"] == "maite"
    modules = set(v["name"] for v in results["modules"])
    # must update this if project's modules are renamed
    assert {
        "maite.testing.docs",
        "maite.testing.project",
        "maite.testing.pyright",
        "maite.testing.pytest",
        "maite.testing",
        "maite.utils.validation",
        "maite.utils",
        "maite",
    } <= modules
    assert not any(
        name.split(".")[-1].startswith("_") for name in modules
    ), "reported module is private"


@pytest.mark.parametrize("submodule", ["", "maite.testing.project"])
def test_public_symbols(submodule):
    symbols = get_public_symbols(
        module_scan("maite"), submodule=submodule, include_dunder_names=False
    )
    names = set(s["name"] for s in symbols)
    assert {
        "maite.testing.project.ModuleScan",
        "maite.testing.project.get_public_symbols",
    } <= names
    for symbol in symbols:
        *_, name = symbol["name"].split(".")

        assert not name.startswith("_"), f"{symbol}: symbol is private"
        assert symbol["isExported"] is True


def test_invalid_submodule():
    results = module_scan("maite")
    with pytest.raises(InvalidArgument, match="11 is not a valid module name."):
        get_public_symbols(results, submodule="maite.11")


def test_special_method_filtering():
    dummy_results = ModuleScanResults(
        version="",
        time="",
        generalDiagnostics=[],
        summary=Summary(
            filesAnalyzed=1,
            errorCount=0,
            warningCount=0,
            informationCount=0,
            timeInSec=0.0,
        ),
        typeCompleteness=CompletenessSection(
            packageName="foo",
            packageRootDirectory=Path("/foo"),
            moduleName="foo",
            moduleRootDirectory=Path("/foo"),
            pyTypedPath=Path("."),
            ignoreUnknownTypesFromImports=False,
            missingClassDocStringCount=0,
            missingDefaultParamCount=0,
            missingFunctionDocStringCount=0,
            exportedSymbolCounts=SymbolCounts(
                withKnownType=5, withAmbiguousType=0, withUnknownType=0
            ),
            otherSymbolCounts=SymbolCounts(
                withKnownType=0, withAmbiguousType=0, withUnknownType=0
            ),
            completenessScore=100,
            modules=["foo"],
            symbols=[
                Symbol(
                    category="class",
                    name="foo.SomeClass",
                    referenceCount=1,
                    isExported=True,
                    isTypeKnown=False,
                    isTypeAmbiguous=False,
                    diagnostics=[],
                ),
                Symbol(
                    category="method",
                    name="foo.SomeClass.__init__",
                    referenceCount=1,
                    isExported=True,
                    isTypeKnown=True,
                    isTypeAmbiguous=False,
                    diagnostics=[],
                ),
                Symbol(
                    category="function",
                    name="foo.SomeClass.__dataclass_method__",
                    referenceCount=1,
                    isExported=True,
                    isTypeKnown=True,
                    isTypeAmbiguous=False,
                    diagnostics=[],
                ),
                Symbol(
                    category="method",
                    name="foo.SomeClass.public_method__",
                    referenceCount=1,
                    isExported=True,
                    isTypeKnown=True,
                    isTypeAmbiguous=False,
                    diagnostics=[],
                ),
                Symbol(
                    category="method",
                    name="foo.SomeClass.___not_special___",
                    referenceCount=1,
                    isExported=True,
                    isTypeKnown=True,
                    isTypeAmbiguous=False,
                    diagnostics=[],
                ),
            ],
        ),
    )
    symbols_with_special = get_public_symbols(dummy_results, include_dunder_names=True)
    symbols_without_special = get_public_symbols(
        dummy_results, include_dunder_names=False
    )

    assert set(x["name"] for x in symbols_with_special) == {
        "foo.SomeClass",
        "foo.SomeClass.__init__",
        "foo.SomeClass.public_method__",
        "foo.SomeClass.___not_special___",
        "foo.SomeClass.__dataclass_method__",
    }

    assert set(x["name"] for x in symbols_without_special) == {
        "foo.SomeClass",
        "foo.SomeClass.public_method__",
        "foo.SomeClass.___not_special___",
    }


def test_import_symbols():
    results = module_scan("maite_dummy.basic")
    out = list(import_public_symbols(results))

    from maite_dummy.basic.stuff import (
        AClass,
        ADataClass,
        AProtocol,
        ATypedDict,
        a_func,
    )

    expected_stuff = {a_func, AClass, ADataClass, AProtocol, ATypedDict}
    assert len(out) == len(expected_stuff)
    assert set(out) == expected_stuff


def test_import_pytest_skip():
    results = module_scan("maite_dummy.basic")
    out: List[str] = [
        x.values[0]
        for x in import_public_symbols(results, skip_module_not_found="pytest-skip")
        if isinstance(x, ParameterSet)
    ]  # type: ignore

    expected = {
        "maite_dummy.basic.needs_mygrad.func_needs_mygrad",
    }
    assert sorted(out) == sorted(expected)


def test_validate_import_public_symbols_input():
    results = module_scan("maite_dummy.basic")

    with pytest.raises(
        InvalidArgument,
        match=r"Expected `skip_module_not_found` to be one of: False, True, pytest-skip. Got `pytest-blah`.",
    ):
        list(import_public_symbols(results, skip_module_not_found="pytest-blah"))  # type: ignore



from __future__ import annotations

from typing import Any, Generic, Iterable, Optional, Sequence, Tuple, TypeVar, overload

from maite._internals.protocols import (
    image_classification as ic,
    object_detection as od,
)
from maite._internals.protocols.generic import (
    Augmentation,
    CollateFn,
    DataLoader,
    Dataset,
    Metric,
    MetricComputeReturnType,
    Model,
)
from maite._internals.protocols.task_aliases import (  # SomeInputType,; SomeMetadataType,; SomeTargetType,
    OpenAugmentationType,
    OpenDataLoaderType,
    OpenDatasetType,
    OpenMetricType,
    OpenModelType,
    SomeInputBatchType,
    SomeInputType,
    SomeMetadataBatchType,
    SomeMetadataType,
    SomeTargetBatchType,
    SomeTargetType,
)
from maite._internals.utils import add_progress_bar
from maite.errors import InvalidArgument

# TODO: Permit returned predictions to be of type tuple[InputDataType, TargetDataType, DatumMetadataType].
#       This seems much more natural as it is independent of batch_size.
#   - This would require we use some method to iterate over "Batch-types" to get individual types
#     and package individual predictions into an iterable. batch objects are iterable and then
#     checking that the type of object returned from the iterable is ArrayLike.)


# TODO: Consider the implications of having a flexible implementation that only relies on membership
#       in generic class types. Can we guarantee (or test) that satisfying one of our overload
#       type signatures is always sufficient to return the types promised by that signature?

#       A: Yes -- This can be done by using static type checker to determine whether evaluate
#          and predict implementations returns promised types from each overload when inputs
#          are typed as promised by call signatures. See gitlab issue 425 and discussion on MR129.


class _DummyMetric(Metric):
    """
    Metric that does nothing and returns an empty dictionary from compute
    """

    def reset(self) -> None:
        ...

    def update(
        self,
        __preds_batch: Any,
        __targets_batch: Any,
    ) -> None:
        return None

    def compute(self) -> MetricComputeReturnType:
        return dict()


# Note: These 3 TypeVars don't capture coupling constraints between
# types of each (i.e. all should be drawn from same subproblem).

T_in = TypeVar("T_in", bound=SomeInputType)
T_tgt = TypeVar("T_tgt", bound=SomeTargetType)
T_md = TypeVar("T_md", bound=SomeMetadataType)


def default_collate_fn(
    batch_data_as_singles: Iterable[Tuple[T_in, T_tgt, T_md]]
) -> Tuple[Sequence[T_in], Sequence[T_tgt], Sequence[T_md]]:
    """
    Describe how to create a tuple of
    (InputBatchType, TargetBatchType, DatumMetadataBatchType)
    from an iterator of tuples of
    (InputType, TargetType, DatumMetadataType)
    """

    # first just unpack inputs/targets/metadata from iterator into separate lists
    input_batch: list[T_in] = []
    target_batch: list[T_tgt] = []
    metadata_batch: list[T_md] = []
    for input_datum, target_datum, metadata_datum in batch_data_as_singles:
        input_batch.append(input_datum)
        target_batch.append(target_datum)
        metadata_batch.append(metadata_datum)

    return input_batch, target_batch, metadata_batch


class _SimpleDataLoader(Generic[T_in, T_tgt, T_md]):
    """
    Simple dataloader to create by default should a user not provide their own
    dataloader but does provide a Dataset.

    Preconditions: dataset and collate_fn come from same ml subproblem domain
    """

    def __init__(
        self,
        dataset: Dataset[T_in, T_tgt, T_md],
        batch_size: int,
        collate_fn: CollateFn = default_collate_fn,
    ):
        self.dataset = dataset
        self.batch_size = batch_size
        self.collate_fn = collate_fn

    def __iter__(self):
        # iterate over first batch_size examples from dataset, collate them, and yield result
        n_batches = (
            len(self.dataset) // self.batch_size
            if len(self.dataset) % self.batch_size == 0
            else len(self.dataset) // self.batch_size + 1
        )

        for batch_no in range(n_batches):
            batch_data_as_singles = [
                self.dataset[i]
                for i in range(
                    batch_no * self.batch_size, (batch_no + 1) * self.batch_size
                )
                if i < len(self.dataset)
            ]

            batch_inputs, batch_targets, batch_metadata = self.collate_fn(
                batch_data_as_singles
            )

            yield (batch_inputs, batch_targets, batch_metadata)


# begin all overloads for evaluate function
# (There are currently no task-specific default arguments, which means
# that all runtime-behavior of evaluate is common and no task-specific
# procedures need to exist, at least not yet.)


@overload
def evaluate(
    *,
    model: ic.Model,
    metric: Optional[ic.Metric] = None,
    dataloader: Optional[ic.DataLoader] = None,
    dataset: Optional[ic.Dataset] = None,
    batch_size: int = 1,
    augmentation: Optional[ic.Augmentation] = None,
    return_augmented_data: bool = False,
    return_preds: bool = False,
) -> Tuple[
    MetricComputeReturnType,
    Sequence[ic.TargetBatchType],
    Sequence[Tuple[ic.InputBatchType, ic.TargetBatchType, ic.DatumMetadataBatchType]],
]:
    ...


@overload
def evaluate(
    *,
    model: od.Model,
    metric: Optional[od.Metric] = None,
    dataloader: Optional[od.DataLoader] = None,
    dataset: Optional[od.Dataset] = None,
    batch_size: int = 1,
    augmentation: Optional[od.Augmentation] = None,
    return_augmented_data: bool = False,
    return_preds: bool = False,
) -> Tuple[
    MetricComputeReturnType,
    Sequence[od.TargetBatchType],
    Sequence[Tuple[od.InputBatchType, od.TargetBatchType, od.DatumMetadataBatchType]],
]:
    ...


def evaluate(
    *,
    model: OpenModelType,
    metric: Optional[OpenMetricType] = None,
    dataloader: Optional[OpenDataLoaderType] = None,
    dataset: Optional[OpenDatasetType] = None,
    batch_size: int = 1,
    augmentation: Optional[OpenAugmentationType] = None,
    return_augmented_data: bool = False,
    return_preds: bool = False,
) -> Tuple[
    MetricComputeReturnType,
    Sequence[SomeTargetBatchType],
    Sequence[Tuple[SomeInputBatchType, SomeTargetBatchType, SomeMetadataBatchType]],
]:
    # doc-ignore: EX01
    """
    Evaluate a model's performance on data according to some metric with optional augmentation.

    Some data source (either a dataloader or a dataset) must be provided
    or an InvalidArgument exception is raised.

    Parameters
    ----------
    model : SomeModel
        Maite Model object.

    metric : Optional[SomeMetric], (default=None)
        Compatible maite Metric.

    dataloader : Optional[SomeDataloader], (default=None)
        Compatible maite dataloader.

    dataset : Optional[SomeDataset], (default=None)
        Compatible maite dataset.

    batch_size : int, (default=1)
        Batch size for use with dataset (ignored if dataset=None).

    augmentation : Optional[SomeAugmentation], (default=None)
        Compatible maite augmentation.

    return_augmented_data : bool, (default=False)
        Set to True to return post-augmentation data as a function output.

    return_preds : bool, (default=False)
        Set to True to return raw predictions as a function output.

    Returns
    -------
    Tuple[Dict[str, Any], Sequence[TargetType], Sequence[Tuple[InputBatchType, TargetBatchType, DatumMetadataBatchType]]]
        Tuple of returned metric value, sequence of model predictions, and
        sequence of data batch tuples fed to the model during inference. The actual
        types represented by InputBatchType, TargetBatchType, and DatumMetadataBatchType will vary
        by the domain of the components provided as input arguments (e.g. image
        classification or object detection.)
        Note that the second and third return arguments will be empty if
        return_augmented_data is False or return_preds is False, respectively.
    """

    return _evaluate(
        model=model,
        metric=metric,
        dataloader=dataloader,
        dataset=dataset,
        batch_size=batch_size,
        augmentation=augmentation,
        collate_fn=default_collate_fn,
        return_augmented_data=return_augmented_data,
        return_preds=return_preds,
    )


def _evaluate(
    *,
    model: Model,
    metric: Optional[Metric] = None,
    dataloader: Optional[DataLoader] = None,
    dataset: Optional[Dataset] = None,
    batch_size: int = 1,
    augmentation: Optional[Augmentation] = None,
    collate_fn: Optional[CollateFn] = None,
    return_augmented_data: bool = False,
    return_preds: bool = False,
) -> Tuple[
    MetricComputeReturnType,
    Sequence[SomeTargetBatchType],
    Sequence[Tuple[SomeInputBatchType, SomeTargetBatchType, SomeMetadataBatchType]],
]:
    """
    Task-agnostically typed version of evaluate for use by any workflows that
    wish to leverage evaluate internally in a task-agnostic way.
    """

    # Validate potential input combinations.
    # User needs to provide ONE of either (a) dataloader (iterable over
    # tuples of batch-types) or (b) a dataset (an iterable over tuples
    # of individual types) and a collate_fn. (batch_size is assumed
    # to be 1 if not provided)

    # if metric is None, make it a _DummyMetric that does nothing

    # if dataloader is None, check that we have Dataset and collate_fn.
    # If so, construct a dataloader, and assign to dataloader variable

    if metric is None:
        # user provided no metric
        metric = _DummyMetric()

    if dataloader is None and dataset is None:
        # user provided neither a dataloader nor a dataset
        raise InvalidArgument("One of dataloader and dataset must be provided")

    if dataloader is None and dataset is not None:
        if collate_fn is None:
            raise InvalidArgument(
                "If dataset is provided, then collate_fn is required"
                + "to permit building a dataloader"
            )

    if dataloader is None:
        assert dataset is not None  # should never trigger due to previous checks
        assert batch_size is not None  # shouldn't trigger due to default value
        assert collate_fn is not None  # should never trigger due to previous checks

        dataloader = _SimpleDataLoader(
            dataset=dataset, batch_size=batch_size, collate_fn=collate_fn
        )

    # dataloader, metric, and model are populated by this point
    assert dataloader is not None
    assert metric is not None
    assert model is not None

    preds_batches = []
    metric.reset()
    augmented_data_batches = []

    dataloader = add_progress_bar(dataloader)

    for input_datum_batch, target_datum_batch, metadata_batch in dataloader:
        if augmentation is not None:
            (
                input_datum_batch_aug,
                target_datum_batch_aug,
                metadata_batch_aug,
            ) = augmentation((input_datum_batch, target_datum_batch, metadata_batch))
        else:
            input_datum_batch_aug, target_datum_batch_aug, metadata_batch_aug = (
                input_datum_batch,
                target_datum_batch,
                metadata_batch,
            )

        preds_batch = model(input_datum_batch_aug)
        metric.update(preds_batch, target_datum_batch_aug)

        # store any requested intermediate data for returning to user
        if return_augmented_data:
            augmented_data_batches.append(
                (input_datum_batch_aug, target_datum_batch_aug, metadata_batch_aug)
            )

        if return_preds:
            preds_batches.append(preds_batch)

    # TODO: investigate better munging/recollating iterable/batches
    # aug_data is currently a sequence of tuples of batch-types
    # this is a little artificial as batch-size is typically less relevant
    # to test ande evaluation and more an implementation detail of the forward
    # pass.

    preds = preds_batches
    aug_data = augmented_data_batches
    metric_results = metric.compute()

    return metric_results, preds, aug_data


@overload
def predict(
    *,
    model: ic.Model,
    dataloader: Optional[ic.DataLoader] = None,
    dataset: Optional[ic.Dataset] = None,
    batch_size: int = 1,
    augmentation: Optional[ic.Augmentation] = None,
) -> Tuple[
    Sequence[ic.TargetBatchType],
    Sequence[Tuple[ic.InputBatchType, ic.TargetBatchType, ic.DatumMetadataBatchType]],
]:
    ...


@overload
def predict(
    *,
    model: od.Model,
    dataloader: Optional[od.DataLoader] = None,
    dataset: Optional[od.Dataset] = None,
    batch_size: int = 1,
    augmentation: Optional[od.Augmentation] = None,
) -> Tuple[
    Sequence[od.TargetBatchType],
    Sequence[Tuple[od.InputBatchType, od.TargetBatchType, od.DatumMetadataBatchType]],
]:
    ...


def predict(
    *,
    model: OpenModelType,
    dataloader: Optional[OpenDataLoaderType] = None,
    dataset: Optional[OpenDatasetType] = None,
    batch_size: int = 1,
    augmentation: Optional[OpenAugmentationType] = None,
) -> Tuple[
    Sequence[SomeTargetBatchType],
    Sequence[Tuple[SomeInputBatchType, SomeTargetBatchType, SomeMetadataBatchType]],
]:
    # doc-ignore: EX01
    """
    Make predictions for a given model & data source with optional augmentation.

    Some data source (either a dataloader or a dataset) must be provided
    or an InvalidArgument exception is raised.

    Parameters
    ----------
    model : SomeModel
        Maite Model object.

    dataloader : Optional[SomeDataloader], (default=None)
        Compatible maite dataloader.

    dataset : Optional[SomeDataset], (default=None)
        Compatible maite dataset.

    batch_size : int, (default=1)
        Batch size for use with dataset (ignored if dataset=None).

    augmentation : Optional[SomeAugmentation], (default=None)
        Compatible maite augmentation.

    Returns
    -------
    Tuple[Sequence[SomeTargetBatchType], Sequence[Tuple[SomeInputBatchType, SomeTargetBatchType, SomeMetadataBatchType]],
        A tuple of the predictions (as a sequence of batches) and a sequence
        of tuples containing the information associated with each batch.
    """

    _, preds, aug_data = _evaluate(
        model=model,
        dataloader=dataloader,
        dataset=dataset,
        augmentation=augmentation,
        batch_size=batch_size,
        collate_fn=default_collate_fn,
        return_augmented_data=True,
        return_preds=True,
    )

    return preds, aug_data



# define fixtures for use by any module in this directory

import pytest

import maite.protocols.image_classification as ic
import maite.protocols.object_detection as od
from tests.component_impls import (
    ic_simple_component_impls as ici,
    od_simple_component_impls as odi,
)


# --- ic component fixtures ---
@pytest.fixture(scope="module")
def ic_simple_dataset() -> ic.Dataset:
    return ici.DatasetImpl()


@pytest.fixture(scope="module")
def ic_simple_dataloader(ic_simple_dataset) -> ic.DataLoader:
    return ici.DataLoaderImpl(dataset=ic_simple_dataset)


@pytest.fixture(scope="module")
def ic_simple_augmentation() -> ic.Augmentation:
    return ici.AugmentationImpl()


@pytest.fixture(scope="module")
def ic_simple_model() -> ic.Model:
    return ici.ModelImpl()


@pytest.fixture(scope="module")
def ic_simple_metric() -> ic.Metric:
    return ici.MetricImpl()


# --- od component fixtures ---
@pytest.fixture(scope="module")
def od_simple_dataset() -> od.Dataset:
    return odi.DatasetImpl()


@pytest.fixture(scope="module")
def od_simple_dataloader(od_simple_dataset) -> od.DataLoader:
    return odi.DataLoaderImpl(dataset=od_simple_dataset)


@pytest.fixture(scope="module")
def od_simple_augmentation() -> od.Augmentation:
    return odi.AugmentationImpl()


@pytest.fixture(scope="module")
def od_simple_model() -> od.Model:
    return odi.ModelImpl()


@pytest.fixture(scope="module")
def od_simple_metric() -> od.Metric:
    return odi.MetricImpl()
